# -*- coding: utf-8 -*-

""" Deeplabv3+ model for Keras.
This model is based on TF repo:
https://github.com/tensorflow/models/tree/master/research/deeplab
On Pascal VOC, original model gets to 84.56% mIOU

Now this model is only available for the TensorFlow backend,
due to its reliance on `SeparableConvolution` layers, but Theano will add
this layer soon.

MobileNetv2 backbone is based on this repo:
https://github.com/JonathanCMitchell/mobilenet_v2_keras

# Reference
- [Encoder-Decoder with Atrous Separable Convolution
    for Semantic Image Segmentation](https://arxiv.org/pdf/1802.02611.pdf)
- [Xception: Deep Learning with Depthwise Separable Convolutions]
    (https://arxiv.org/abs/1610.02357)
- [Inverted Residuals and Linear Bottlenecks: Mobile Networks for
    Classification, Detection and Segmentation](https://arxiv.org/abs/1801.04381)
"""
from keras import backend as K
from keras.layers import InputSpec, Layer
from keras.utils import conv_utils


class BilinearUpsampling(Layer):
    """Just a simple bilinear upsampling layer. Works only with TF.
    Args:
        upsampling: tuple of 2 numbers > 0. The upsampling ratio for h and w
        output_size: used instead of upsampling arg if passed!
    """

    def __init__(self, upsampling=(2, 2), output_size=None, data_format=None, **kwargs):

        super(BilinearUpsampling, self).__init__(**kwargs)

        self.data_format = conv_utils.normalize_data_format(data_format)
        self.input_spec = InputSpec(ndim=4)
        if output_size:
            self.output_size = conv_utils.normalize_tuple(output_size, 2, "output_size")
            self.upsampling = None
        else:
            self.output_size = None
            self.upsampling = conv_utils.normalize_tuple(upsampling, 2, "upsampling")

    def compute_output_shape(self, input_shape):
        if self.upsampling:
            height = self.upsampling[0] * input_shape[1] if input_shape[1] is not None else None
            width = self.upsampling[1] * input_shape[2] if input_shape[2] is not None else None
        else:
            height = self.output_size[0]
            width = self.output_size[1]
        return (input_shape[0], height, width, input_shape[3])

    def call(self, inputs):
        if self.upsampling:
            return K.tf.image.resize_bilinear(
                inputs,
                (inputs.shape[1] * self.upsampling[0], inputs.shape[2] * self.upsampling[1]),
                align_corners=True,
            )
        else:
            return K.tf.image.resize_bilinear(
                inputs, (self.output_size[0], self.output_size[1]), align_corners=True
            )

    def get_config(self):
        config = {
            "upsampling": self.upsampling,
            "output_size": self.output_size,
            "data_format": self.data_format,
        }
        base_config = super(BilinearUpsampling, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



from typing import Sequence

from fvcore.common.registry import Registry

from medsegpy.config import Config
from medsegpy.data import MetadataCatalog

EVALUATOR_REGISTRY = Registry("EVALUATOR")
"""
Registry for evaluators, which process input/output pairs to compute metric
scores and save data to disk. The evaluator type should be registered with
it's metadata in :class:`MetadataCatalog`.

The registered object will be called with
`obj(dataset_name, cfg, output_dir, save_raw_data)`.
The call should return a :class:`DatasetEvaluator` object.
"""


def build_evaluator(
    dataset_name: str, cfg: Config, output_folder: str = None, save_raw_data: bool = False
):
    name = MetadataCatalog.get(dataset_name).evaluator_type
    if isinstance(name, str):
        name = [name]
    elif isinstance(name, dict):
        primary_task = cfg.PRIMARY_TASK
        assert primary_task in name, (
            f"Primary Task (= {primary_task}) is not a key in "
            f"evaluator dictionary for dataset '{dataset_name}'"
        )
        name = [name[primary_task]]
    assert isinstance(name, Sequence)

    evaluators = []
    for n in name:
        evaluator_cls = EVALUATOR_REGISTRY.get(n)
        evaluators.append(evaluator_cls(dataset_name, cfg, output_folder, save_raw_data))

    return evaluators



import itertools
import unittest

from medsegpy.data.data_utils import compute_patches


class TestComputePatches(unittest.TestCase):
    def verify_expected_patches(self, patches, expected):
        """Verify patches are in expected patches based on"""
        pass

    def test_simple(self):
        image_size = (10, 20, 30)

        # Patch size is whole volume.
        patches = compute_patches(image_size, image_size)
        assert len(patches) == 1
        patch, pad = patches[0]
        assert patch == (slice(0, 10), slice(0, 20), slice(0, 30))
        assert pad is None

        # Patch size is uniformly 1/2 on all dimensions.
        patch_size = (5, 10, 15)
        patches = compute_patches(image_size, patch_size)
        assert len(patches) == 2 * 2 * 2
        expected_patches = itertools.product(
            (slice(0, 5), slice(5, 10)),
            (slice(0, 10), slice(10, 20)),
            (slice(0, 15), slice(15, 30)),
        )
        assert all(patch in expected_patches for patch, _ in patches)
        assert all(pad is None for _, pad in patches)

        # Patch size is different on all dimensions.
        patch_size = (2, 10, 5)
        patches = compute_patches(image_size, patch_size)
        assert len(patches) == 5 * 2 * 6
        expected_patches = itertools.product(
            (slice(0, 2), slice(2, 4), slice(4, 6), slice(6, 8), slice(8, 10)),
            (slice(0, 10), slice(10, 20)),
            (
                slice(0, 5),
                slice(5, 10),
                slice(10, 15),
                slice(15, 20),
                slice(20, 25),
                slice(25, 30),
            ),  # noqa
        )
        assert all(patch in expected_patches for patch, _ in patches)
        assert all(pad is None for _, pad in patches)

    def test_padding(self):
        image_size = (10, 20, 30)
        patch_size = (5, 10, 15)

        # Pad with all dimensions with padding 1.
        patches = compute_patches(image_size, patch_size, pad_size=1)
        expected_patches = itertools.product(
            (slice(0, 4), slice(4, 9)), (slice(0, 9), slice(9, 19)), (slice(0, 14), slice(14, 29))
        )
        assert all(patch in expected_patches for patch, _ in patches)
        expected_pads = itertools.product(((1, 0), (0, 0)), ((1, 0), (0, 0)), ((1, 0), (0, 0)))
        expected_pads = tuple(None if all(px == (0, 0) for px in p) else p for p in expected_pads)
        assert tuple(pad for _, pad in patches) == expected_pads

        # Pad only first dimension
        patches = compute_patches(image_size, patch_size, pad_size=(1, None, None))
        expected_patches = itertools.product(
            (slice(0, 4), slice(4, 9)), (slice(0, 10), slice(10, 20)), (slice(0, 15), slice(15, 30))
        )
        assert all(patch in expected_patches for patch, _ in patches)
        expected_pads = itertools.product(((1, 0), (0, 0)), ((0, 0), (0, 0)), ((0, 0), (0, 0)))
        expected_pads = tuple(None if all(px == (0, 0) for px in p) else p for p in expected_pads)
        assert tuple(pad for _, pad in patches) == expected_pads

        # Pad such that dimensions have padding on both front and end
        patches = compute_patches(image_size, patch_size, pad_size=(3, 5, 8))
        expected_patches = itertools.product(
            (slice(0, 2), slice(2, 7), slice(7, 10)),
            (slice(0, 5), slice(5, 15), slice(15, 20)),
            (slice(0, 7), slice(7, 22), slice(22, 30)),
        )
        assert all(patch in expected_patches for patch, _ in patches)
        expected_pads = itertools.product(
            ((3, 0), (0, 0), (0, 2)), ((5, 0), (0, 0), (0, 5)), ((8, 0), (0, 0), (0, 7))
        )
        expected_pads = tuple(None if all(px == (0, 0) for px in p) else p for p in expected_pads)
        assert tuple(pad for _, pad in patches) == expected_pads

    def test_stride(self):
        image_size = (10, 20, 30)
        patch_size = (5, 10, 15)

        # Variable strides without padding.
        patches = compute_patches(image_size, patch_size, strides=(2, 4, 6))
        expected_patches = itertools.product(
            (slice(0, 5), slice(2, 7), slice(4, 9)),
            (slice(0, 10), slice(4, 14), slice(8, 18)),
            (slice(0, 15), slice(6, 21), slice(12, 27)),
        )
        assert all(patch in expected_patches for patch, _ in patches)
        assert all(pad is None for _, pad in patches)

        # Variable strides with padding.
        patches = compute_patches(image_size, patch_size, pad_size=(3, 5, 8), strides=(2, 4, 6))
        expected_patches = list(
            itertools.product(
                (
                    slice(0, 2),
                    slice(0, 4),
                    slice(1, 6),
                    slice(3, 8),
                    slice(5, 10),
                    slice(7, 10),
                ),  # noqa
                (
                    slice(0, 5),
                    slice(0, 9),
                    slice(3, 13),
                    slice(7, 17),
                    slice(11, 20),
                    slice(15, 20),
                ),  # noqa
                (
                    slice(0, 7),
                    slice(0, 13),
                    slice(4, 19),
                    slice(10, 25),
                    slice(16, 30),
                    slice(22, 30),
                ),  # noqa
            )
        )
        assert all(patch in expected_patches for patch, _ in patches)
        expected_pads = itertools.product(
            ((3, 0), (1, 0), (0, 0), (0, 0), (0, 0), (0, 2)),
            ((5, 0), (1, 0), (0, 0), (0, 0), (0, 1), (0, 5)),
            ((8, 0), (2, 0), (0, 0), (0, 0), (0, 1), (0, 7)),
        )
        expected_pads = tuple(None if all(px == (0, 0) for px in p) else p for p in expected_pads)
        assert tuple(pad for _, pad in patches) == expected_pads


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

""" Deeplabv3+ model for Keras.
This model is based on TF repo:
https://github.com/tensorflow/models/tree/master/research/deeplab
On Pascal VOC, original model gets to 84.56% mIOU

Now this model is only available for the TensorFlow backend,
due to its reliance on `SeparableConvolution` layers, but Theano will add
this layer soon.

MobileNetv2 backbone is based on this repo:
https://github.com/JonathanCMitchell/mobilenet_v2_keras

# Reference
- [Encoder-Decoder with Atrous Separable Convolution
    for Semantic Image Segmentation](https://arxiv.org/pdf/1802.02611.pdf)
- [Xception: Deep Learning with Depthwise Separable Convolutions]
    (https://arxiv.org/abs/1610.02357)
- [Inverted Residuals and Linear Bottlenecks: Mobile Networks for
    Classification, Detection and Segmentation](https://arxiv.org/abs/1801.04381)
"""
import logging
import warnings

import numpy as np
from keras import backend as K
from keras import layers
from keras.applications import imagenet_utils
from keras.layers import (
    Activation,
    Add,
    AveragePooling2D,
    BatchNormalization,
    Concatenate,
    Conv2D,
    DepthwiseConv2D,
    Dropout,
    Input,
    ZeroPadding2D,
)
from keras.utils.data_utils import get_file
from keras.utils.layer_utils import get_source_inputs

from medsegpy.modeling import initializer_utils as init_utils

from ..layers import BilinearUpsampling
from ..model import Model

logger = logging.getLogger(__name__)

WEIGHTS_PATH_X = "https://github.com/bonlime/keras-deeplab-v3-plus/releases/download/1.1/deeplabv3_xception_tf_dim_ordering_tf_kernels.h5"  # noqa: E501
WEIGHTS_PATH_MOBILE = "https://github.com/bonlime/keras-deeplab-v3-plus/releases/download/1.1/deeplabv3_mobilenetv2_tf_dim_ordering_tf_kernels.h5"  # noqa: E501


class DeeplabModel:
    def __init__(self, kernel_initializer="", seed=None):
        warnings.warn(
            "DeeplabModel is deprecated and will be removed in v0.0.2. "
            "Use `DeeplabV3Plus` instead.",
            DeprecationWarning,
        )
        self.kernel_initializer = kernel_initializer
        self.seed = seed

    def SepConv_BN(
        self,
        x,
        filters,
        prefix,
        stride=1,
        kernel_size=3,
        rate=1,
        depth_activation=False,
        epsilon=1e-3,
    ):
        """SepConv with BN between depthwise & pointwise. Optionally add activation after BN
        Implements right "same" padding for even kernel sizes
        Args:
            x: input tensor
            filters: num of filters in pointwise convolution
            prefix: prefix before name
            stride: stride at depthwise conv
            kernel_size: kernel size for depthwise convolution
            rate: atrous rate for depthwise convolution
            depth_activation: flag to use activation between depthwise & poinwise convs
            epsilon: epsilon to use in BN layer
        """
        if stride == 1:
            depth_padding = "same"
        else:
            kernel_size_effective = kernel_size + (kernel_size - 1) * (rate - 1)
            pad_total = kernel_size_effective - 1
            pad_beg = pad_total // 2
            pad_end = pad_total - pad_beg
            x = ZeroPadding2D((pad_beg, pad_end))(x)
            depth_padding = "valid"

        if not depth_activation:
            x = Activation("relu")(x)
        x = DepthwiseConv2D(
            (kernel_size, kernel_size),
            strides=(stride, stride),
            dilation_rate=(rate, rate),
            padding=depth_padding,
            use_bias=False,
            kernel_initializer=init_utils.get_initializer(self.kernel_initializer, self.seed),
            name=prefix + "_depthwise",
        )(x)
        x = BatchNormalization(name=prefix + "_depthwise_BN", epsilon=epsilon)(x)
        if depth_activation:
            x = Activation("relu")(x)
        x = Conv2D(
            filters,
            (1, 1),
            padding="same",
            use_bias=False,
            kernel_initializer=init_utils.get_initializer(self.kernel_initializer, self.seed),
            name=prefix + "_pointwise",
        )(x)
        x = BatchNormalization(name=prefix + "_pointwise_BN", epsilon=epsilon)(x)
        if depth_activation:
            x = Activation("relu")(x)

        return x

    def _conv2d_same(self, x, filters, prefix, stride=1, kernel_size=3, rate=1):
        """Implements right 'same' padding for even kernel sizes
        Without this there is a 1 pixel drift when stride = 2
        Args:
            x: input tensor
            filters: num of filters in pointwise convolution
            prefix: prefix before name
            stride: stride at depthwise conv
            kernel_size: kernel size for depthwise convolution
            rate: atrous rate for depthwise convolution
        """
        if stride == 1:
            return Conv2D(
                filters,
                (kernel_size, kernel_size),
                strides=(stride, stride),
                padding="same",
                use_bias=False,
                dilation_rate=(rate, rate),
                kernel_initializer=init_utils.get_initializer(self.kernel_initializer, self.seed),
                name=prefix,
            )(x)
        else:
            kernel_size_effective = kernel_size + (kernel_size - 1) * (rate - 1)
            pad_total = kernel_size_effective - 1
            pad_beg = pad_total // 2
            pad_end = pad_total - pad_beg
            x = ZeroPadding2D((pad_beg, pad_end))(x)
            return Conv2D(
                filters,
                (kernel_size, kernel_size),
                strides=(stride, stride),
                padding="valid",
                use_bias=False,
                dilation_rate=(rate, rate),
                kernel_initializer=init_utils.get_initializer(self.kernel_initializer, self.seed),
                name=prefix,
            )(x)

    def _xception_block(
        self,
        inputs,
        depth_list,
        prefix,
        skip_connection_type,
        stride,
        rate=1,
        depth_activation=False,
        return_skip=False,
    ):
        """Basic building block of modified Xception network
        Args:
            inputs: input tensor
            depth_list: number of filters in each SepConv layer. len(depth_list) == 3
            prefix: prefix before name
            skip_connection_type: one of {'conv','sum','none'}
            stride: stride at last depthwise conv
            rate: atrous rate for depthwise convolution
            depth_activation: flag to use activation between depthwise & pointwise convs
            return_skip: flag to return additional tensor after 2 SepConvs for decoder
        """
        residual = inputs
        for i in range(3):
            residual = self.SepConv_BN(
                residual,
                depth_list[i],
                prefix + "_separable_conv{}".format(i + 1),
                stride=stride if i == 2 else 1,
                rate=rate,
                depth_activation=depth_activation,
            )
            if i == 1:
                skip = residual
        if skip_connection_type == "conv":
            shortcut = self._conv2d_same(
                inputs, depth_list[-1], prefix + "_shortcut", kernel_size=1, stride=stride
            )
            shortcut = BatchNormalization(name=prefix + "_shortcut_BN")(shortcut)
            outputs = layers.add([residual, shortcut])
        elif skip_connection_type == "sum":
            outputs = layers.add([residual, inputs])
        elif skip_connection_type == "none":
            outputs = residual
        if return_skip:
            return outputs, skip
        else:
            return outputs

    def _make_divisible(self, v, divisor, min_value=None):
        if min_value is None:
            min_value = divisor
        new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
        # Make sure that round down does not go down by more than 10%.
        if new_v < 0.9 * v:
            new_v += divisor
        return new_v

    def _inverted_res_block(
        self, inputs, expansion, stride, alpha, filters, block_id, skip_connection, rate=1
    ):
        in_channels = inputs._keras_shape[-1]
        pointwise_conv_filters = int(filters * alpha)
        pointwise_filters = self._make_divisible(pointwise_conv_filters, 8)
        x = inputs
        prefix = "expanded_conv_{}_".format(block_id)
        if block_id:
            # Expand

            x = Conv2D(
                expansion * in_channels,
                kernel_size=1,
                padding="same",
                use_bias=False,
                activation=None,
                kernel_initializer=init_utils.get_initializer(self.kernel_initializer, self.seed),
                name=prefix + "expand",
            )(x)
            x = BatchNormalization(epsilon=1e-3, momentum=0.999, name=prefix + "expand_BN")(x)
            x = Activation(self.relu6, name=prefix + "expand_relu")(x)
        else:
            prefix = "expanded_conv_"
        # Depthwise
        x = DepthwiseConv2D(
            kernel_size=3,
            strides=stride,
            activation=None,
            use_bias=False,
            padding="same",
            dilation_rate=(rate, rate),
            kernel_initializer=init_utils.get_initializer(self.kernel_initializer, self.seed),
            name=prefix + "depthwise",
        )(x)
        x = BatchNormalization(epsilon=1e-3, momentum=0.999, name=prefix + "depthwise_BN")(x)

        x = Activation(self.relu6, name=prefix + "depthwise_relu")(x)

        # Project
        x = Conv2D(
            pointwise_filters,
            kernel_size=1,
            padding="same",
            use_bias=False,
            activation=None,
            kernel_initializer=init_utils.get_initializer(self.kernel_initializer, self.seed),
            name=prefix + "project",
        )(x)
        x = BatchNormalization(epsilon=1e-3, momentum=0.999, name=prefix + "project_BN")(x)

        if skip_connection:
            return Add(name=prefix + "add")([inputs, x])

        # if in_channels == pointwise_filters and stride == 1:
        #    return Add(name='res_connect_' + str(block_id))([inputs, x])

        return x

    def Deeplabv3(
        self,
        weights="pascal_voc",
        input_tensor=None,
        input_shape=(512, 512, 3),
        classes=21,
        backbone="mobilenetv2",
        OS=16,
        alpha=1.0,
        dilation_divisor=1,
        dil_rate_input=None,
        dropout_rate=0.1,
    ):
        """Instantiates the Deeplabv3+ architecture

        Optionally loads weights pre-trained
        on PASCAL VOC. This model is available for TensorFlow only,
        and can only be used with inputs following the TensorFlow
        data format `(width, height, channels)`.
        # Arguments
            weights: one of 'pascal_voc' (pre-trained on pascal voc)
                or None (random initialization)
            input_tensor: optional Keras tensor (i.e. output of `layers.Input()`)
                to use as image input for the model.
            input_shape: shape of input image. format HxWxC
                PASCAL VOC model was trained on (512,512,3) images
            classes: number of desired classes. If classes != 21,
                last layer is initialized randomly
            backbone: backbone to use. one of {'xception','mobilenetv2'}
            OS: determines input_shape/feature_extractor_output ratio. One of {8,16}.
                Used only for xception backbone.
            alpha: controls the width of the MobileNetV2 network. This is known as the
                width multiplier in the MobileNetV2 paper.
                    - If `alpha` < 1.0, proportionally decreases the number
                        of filters in each layer.
                    - If `alpha` > 1.0, proportionally increases the number
                        of filters in each layer.
                    - If `alpha` = 1, default number of filters from the paper
                        are used at each layer.
                Used only for mobilenetv2 backbone

        # Returns
            A Keras model instance.

        # Raises
            RuntimeError: If attempting to run this model with a
                backend that does not support separable convolutions.
            ValueError: in case of invalid argument for `weights` or `backbone`

        """
        logger.info("Initializing deeplab model")

        if not (weights in {"pascal_voc", None}):
            raise ValueError(
                "The `weights` argument should be either "
                "`None` (random initialization) or `pascal_voc` "
                "(pre-trained on PASCAL VOC)"
            )

        if K.backend() != "tensorflow":
            raise RuntimeError(
                "The Deeplabv3+ model is only available with " "the TensorFlow backend."
            )

        if not (backbone in {"xception", "mobilenetv2"}):
            raise ValueError(
                "The `backbone` argument should be either " "`xception`  or `mobilenetv2` "
            )

        if input_tensor is None:
            img_input = Input(shape=input_shape)
        else:
            if not K.is_keras_tensor(input_tensor):
                img_input = Input(tensor=input_tensor, shape=input_shape)
            else:
                img_input = input_tensor

        if backbone == "xception":
            if OS == 8:
                entry_block3_stride = 1
                middle_block_rate = 2  # ! Not mentioned in paper, but required
                exit_block_rates = (2, 4)
                atrous_rates = (12, 24, 36)
            else:
                entry_block3_stride = 2
                middle_block_rate = 1
                exit_block_rates = (1, 2)
                atrous_rates = (
                    6 // dilation_divisor,
                    12 // dilation_divisor,
                    18 // dilation_divisor,
                )

            if dil_rate_input is not None:
                atrous_rates = dil_rate_input

            x = Conv2D(
                32,
                (3, 3),
                strides=(2, 2),
                name="entry_flow_conv1_1",
                use_bias=False,
                padding="same",
                kernel_initializer=init_utils.get_initializer(self.kernel_initializer, self.seed),
            )(img_input)
            x = BatchNormalization(name="entry_flow_conv1_1_BN")(x)
            x = Activation("relu")(x)

            x = self._conv2d_same(x, 64, "entry_flow_conv1_2", kernel_size=3, stride=1)
            x = BatchNormalization(name="entry_flow_conv1_2_BN")(x)
            x = Activation("relu")(x)

            x = self._xception_block(
                x,
                [128, 128, 128],
                "entry_flow_block1",
                skip_connection_type="conv",
                stride=2,
                depth_activation=False,
            )
            x, skip1 = self._xception_block(
                x,
                [256, 256, 256],
                "entry_flow_block2",
                skip_connection_type="conv",
                stride=2,
                depth_activation=False,
                return_skip=True,
            )

            x = self._xception_block(
                x,
                [728, 728, 728],
                "entry_flow_block3",
                skip_connection_type="conv",
                stride=entry_block3_stride,
                depth_activation=False,
            )
            for i in range(16):
                x = self._xception_block(
                    x,
                    [728, 728, 728],
                    "middle_flow_unit_{}".format(i + 1),
                    skip_connection_type="sum",
                    stride=1,
                    rate=middle_block_rate,
                    depth_activation=False,
                )

            x = self._xception_block(
                x,
                [728, 1024, 1024],
                "exit_flow_block1",
                skip_connection_type="conv",
                stride=1,
                rate=exit_block_rates[0],
                depth_activation=False,
            )
            x = self._xception_block(
                x,
                [1536, 1536, 2048],
                "exit_flow_block2",
                skip_connection_type="none",
                stride=1,
                rate=exit_block_rates[1],
                depth_activation=True,
            )

        else:
            OS = 8
            first_block_filters = self._make_divisible(32 * alpha, 8)
            x = Conv2D(
                first_block_filters,
                kernel_size=3,
                strides=(2, 2),
                padding="same",
                use_bias=False,
                name="Conv",
                kernel_initializer=init_utils.get_initializer(self.kernel_initializer, self.seed),
            )(img_input)
            x = BatchNormalization(epsilon=1e-3, momentum=0.999, name="Conv_BN")(x)
            x = Activation(self.relu6, name="Conv_Relu6")(x)

            x = self._inverted_res_block(
                x, filters=16, alpha=alpha, stride=1, expansion=1, block_id=0, skip_connection=False
            )

            x = self._inverted_res_block(
                x, filters=24, alpha=alpha, stride=2, expansion=6, block_id=1, skip_connection=False
            )
            x = self._inverted_res_block(
                x, filters=24, alpha=alpha, stride=1, expansion=6, block_id=2, skip_connection=True
            )

            x = self._inverted_res_block(
                x, filters=32, alpha=alpha, stride=2, expansion=6, block_id=3, skip_connection=False
            )
            x = self._inverted_res_block(
                x, filters=32, alpha=alpha, stride=1, expansion=6, block_id=4, skip_connection=True
            )
            x = self._inverted_res_block(
                x, filters=32, alpha=alpha, stride=1, expansion=6, block_id=5, skip_connection=True
            )

            # stride in block 6 changed from 2 -> 1, so we need to use rate = 2
            x = self._inverted_res_block(
                x,
                filters=64,
                alpha=alpha,
                stride=1,  # 1!
                expansion=6,
                block_id=6,
                skip_connection=False,
            )
            x = self._inverted_res_block(
                x,
                filters=64,
                alpha=alpha,
                stride=1,
                rate=2,
                expansion=6,
                block_id=7,
                skip_connection=True,
            )
            x = self._inverted_res_block(
                x,
                filters=64,
                alpha=alpha,
                stride=1,
                rate=2,
                expansion=6,
                block_id=8,
                skip_connection=True,
            )
            x = self._inverted_res_block(
                x,
                filters=64,
                alpha=alpha,
                stride=1,
                rate=2,
                expansion=6,
                block_id=9,
                skip_connection=True,
            )

            x = self._inverted_res_block(
                x,
                filters=96,
                alpha=alpha,
                stride=1,
                rate=2,
                expansion=6,
                block_id=10,
                skip_connection=False,
            )
            x = self._inverted_res_block(
                x,
                filters=96,
                alpha=alpha,
                stride=1,
                rate=2,
                expansion=6,
                block_id=11,
                skip_connection=True,
            )
            x = self._inverted_res_block(
                x,
                filters=96,
                alpha=alpha,
                stride=1,
                rate=2,
                expansion=6,
                block_id=12,
                skip_connection=True,
            )

            x = self._inverted_res_block(
                x,
                filters=160,
                alpha=alpha,
                stride=1,
                rate=2,  # 1!
                expansion=6,
                block_id=13,
                skip_connection=False,
            )
            x = self._inverted_res_block(
                x,
                filters=160,
                alpha=alpha,
                stride=1,
                rate=4,
                expansion=6,
                block_id=14,
                skip_connection=True,
            )
            x = self._inverted_res_block(
                x,
                filters=160,
                alpha=alpha,
                stride=1,
                rate=4,
                expansion=6,
                block_id=15,
                skip_connection=True,
            )

            x = self._inverted_res_block(
                x,
                filters=320,
                alpha=alpha,
                stride=1,
                rate=4,
                expansion=6,
                block_id=16,
                skip_connection=False,
            )

        # end of feature extractor

        # branching for Atrous Spatial Pyramid Pooling

        # Image Feature branch
        # out_shape = int(np.ceil(input_shape[0] / OS))
        b4 = AveragePooling2D(
            pool_size=(int(np.ceil(input_shape[0] / OS)), int(np.ceil(input_shape[1] / OS)))
        )(x)
        b4 = Conv2D(
            256,
            (1, 1),
            padding="same",
            use_bias=False,
            kernel_initializer=init_utils.get_initializer(self.kernel_initializer, self.seed),
            name="image_pooling",
        )(b4)
        b4 = BatchNormalization(name="image_pooling_BN", epsilon=1e-5)(b4)
        b4 = Activation("relu")(b4)
        b4 = BilinearUpsampling(
            (int(np.ceil(input_shape[0] / OS)), int(np.ceil(input_shape[1] / OS)))
        )(b4)

        # simple 1x1
        b0 = Conv2D(
            256,
            (1, 1),
            padding="same",
            use_bias=False,
            kernel_initializer=init_utils.get_initializer(self.kernel_initializer, self.seed),
            name="aspp0",
        )(x)
        b0 = BatchNormalization(name="aspp0_BN", epsilon=1e-5)(b0)
        b0 = Activation("relu", name="aspp0_activation")(b0)

        # there are only 2 branches in mobilenetV2. not sure why
        if backbone == "xception":
            # rate = 6 (12)
            b1 = self.SepConv_BN(
                x, 256, "aspp1", rate=atrous_rates[0], depth_activation=True, epsilon=1e-5
            )
            # rate = 12 (24)
            b2 = self.SepConv_BN(
                x, 256, "aspp2", rate=atrous_rates[1], depth_activation=True, epsilon=1e-5
            )
            # rate = 18 (36)
            b3 = self.SepConv_BN(
                x, 256, "aspp3", rate=atrous_rates[2], depth_activation=True, epsilon=1e-5
            )

            # concatenate ASPP branches & project
            x = Concatenate()([b4, b0, b1, b2, b3])
        else:
            x = Concatenate()([b4, b0])

        x = Conv2D(
            256,
            (1, 1),
            padding="same",
            use_bias=False,
            kernel_initializer=init_utils.get_initializer(self.kernel_initializer, self.seed),
            name="concat_projection",
        )(x)
        x = BatchNormalization(name="concat_projection_BN", epsilon=1e-5)(x)
        x = Activation("relu")(x)
        x = Dropout(dropout_rate, seed=self.seed)(x)

        # DeepLab v.3+ decoder

        if backbone == "xception":
            # Feature projection
            # x4 (x2) block
            x = BilinearUpsampling(
                output_size=(int(np.ceil(input_shape[0] / 4)), int(np.ceil(input_shape[1] / 4)))
            )(x)
            dec_skip1 = Conv2D(
                48,
                (1, 1),
                padding="same",
                use_bias=False,
                kernel_initializer=init_utils.get_initializer(self.kernel_initializer, self.seed),
                name="feature_projection0",
            )(skip1)
            dec_skip1 = BatchNormalization(name="feature_projection0_BN", epsilon=1e-5)(dec_skip1)
            dec_skip1 = Activation("relu")(dec_skip1)
            x = Concatenate()([x, dec_skip1])
            x = self.SepConv_BN(x, 256, "decoder_conv0", depth_activation=True, epsilon=1e-5)
            x = self.SepConv_BN(x, 256, "decoder_conv1", depth_activation=True, epsilon=1e-5)

        # you can use it with arbitary number of classes
        if classes == 21:
            last_layer_name = "logits_semantic"
        else:
            last_layer_name = "custom_logits_semantic"

        x = Conv2D(
            classes,
            (1, 1),
            padding="same",
            kernel_initializer=init_utils.get_initializer(self.kernel_initializer, self.seed),
            name=last_layer_name,
        )(x)
        x = BilinearUpsampling(output_size=(input_shape[0], input_shape[1]))(x)

        # Ensure that the model takes into account
        # any potential predecessors of `input_tensor`.
        if input_tensor is not None:
            inputs = get_source_inputs(input_tensor)
        else:
            inputs = img_input

        model = Model(inputs, x, name="deeplabv3+")

        # load weights

        if weights == "pascal_voc":
            if backbone == "xception":
                weights_path = get_file(
                    "deeplabv3_xception_tf_dim_ordering_tf_kernels.h5",
                    WEIGHTS_PATH_X,
                    cache_subdir="models",
                )
            else:
                weights_path = get_file(
                    "deeplabv3_mobilenetv2_tf_dim_ordering_tf_kernels.h5",
                    WEIGHTS_PATH_MOBILE,
                    cache_subdir="models",
                )
            model.load_weights(weights_path, by_name=True)
        return model

    def preprocess_input(self, x):
        """Preprocesses a numpy array encoding a batch of images.
        # Arguments
            x: a 4D numpy array consists of RGB values within [0, 255].
        # Returns
            Input array scaled to [-1.,1.]
        """
        return imagenet_utils.preprocess_input(x, mode="tf")

    def relu6(self, x):
        return K.relu(x, max_value=6)



from .build import META_ARCH_REGISTRY, build_model
from .deeplabv3 import DeeplabV3Plus
from .fc_densenet import FCDenseNet
from .inpainting_and_seg_unet import ContextEncoder
from .unet import UNet2D, UNet3D, build_decoder_block, build_encoder_block



"""Build dataset dictionaries."""
import itertools
import logging
from typing import Dict, Hashable, List, Sequence, Union

from medsegpy.config import Config

from .catalog import DatasetCatalog
from .data_loader import build_data_loader


def filter_dataset(
    dataset_dicts: List[Dict], by: Hashable, accepted_elements, include_missing: bool = False
):
    """Filter by common dataset fields.

    Args:
        dataset_dicts (List[Dict]): data in MedSegPy Dataset format.
        by (Hashable): Field to filter by.
        accepted_elements (Sequence): Acceptable elements.
        include_missing (bool, optional): If `True`, include elements without
            `by` field in dictionary representation.

    Returns:
        List[Dict]: Filtered dataset dictionaries.
    """
    num_before = len(dataset_dicts)
    dataset_dicts = [
        x for x in dataset_dicts if include_missing or (by in x and x[by] in accepted_elements)
    ]
    num_after = len(dataset_dicts)

    logger = logging.getLogger(__name__)
    logger.info(
        "Removed {} elements with filter '{}'. {} elements left.".format(
            num_before - num_after, by, num_after
        )
    )
    return dataset_dicts


def get_sem_seg_dataset_dicts(dataset_names: Sequence[str], filter_empty: bool = True):
    """Load and prepare dataset dicts for semantic segmentation.

    Args:
        dataset_names (Sequence[str])": A list of dataset names.
        filter_empty (bool, optional): Filter datasets without field
            `'sem_seg_file'`.
    """
    assert len(dataset_names)
    dataset_dicts = [DatasetCatalog.get(dataset_name) for dataset_name in dataset_names]
    for dataset_name, dicts in zip(dataset_names, dataset_dicts):
        assert len(dicts), "Dataset '{}' is empty!".format(dataset_name)

    dataset_dicts = list(itertools.chain.from_iterable(dataset_dicts))

    logger = logging.getLogger(__name__)
    if filter_empty:
        num_before = len(dataset_dicts)
        dataset_dicts = [x for x in dataset_dicts if "sem_seg_file" in x]
        num_after = len(dataset_dicts)
        logger.info(
            "Removed {} elements without annotations. {} elements left.".format(
                num_before - num_after, num_after
            )
        )

    return dataset_dicts


def build_loader(
    cfg: Config,
    dataset_names: Union[str, Sequence[str]],
    batch_size: int,
    is_test: bool,
    shuffle: bool,
    drop_last: bool,
    **kwargs,
):
    if isinstance(dataset_names, str):
        dataset_names = (dataset_names,)
    kwargs["batch_size"] = batch_size
    kwargs["is_test"] = is_test
    kwargs["shuffle"] = shuffle
    kwargs["drop_last"] = drop_last

    remove_files_without_seg = True
    # If the task is inpainting, we want to keep all unlabeled data
    # because inpainting does not require labels
    if cfg.PRIMARY_TASK == "inpainting":
        remove_files_without_seg = False

    dataset_dicts = get_sem_seg_dataset_dicts(dataset_names, filter_empty=remove_files_without_seg)
    return build_data_loader(cfg, dataset_dicts, **kwargs)



"""Adopted from https://github.com/keras-team/keras/issues/3556#issuecomment-440638517"""  # noqa

# TODO (TF2.X)
# from keras.legacy import interfaces
import keras.backend as K
from keras.optimizers import Optimizer


class AdamAccumulate(Optimizer):
    def __init__(
        self,
        lr=0.001,
        beta_1=0.9,
        beta_2=0.999,
        epsilon=None,
        decay=0.0,
        amsgrad=False,
        accum_iters=1,
        **kwargs
    ):
        if accum_iters < 1:
            raise ValueError("accum_iters must be >= 1")
        super(AdamAccumulate, self).__init__(**kwargs)
        with K.name_scope(self.__class__.__name__):
            self.iterations = K.variable(0, dtype="int64", name="iterations")
            self.lr = K.variable(lr, name="lr")
            self.beta_1 = K.variable(beta_1, name="beta_1")
            self.beta_2 = K.variable(beta_2, name="beta_2")
            self.decay = K.variable(decay, name="decay")
        if epsilon is None:
            epsilon = K.epsilon()
        self.epsilon = epsilon
        self.initial_decay = decay
        self.amsgrad = amsgrad
        self.accum_iters = K.variable(accum_iters, K.dtype(self.iterations))
        self.accum_iters_float = K.cast(self.accum_iters, K.floatx())

    # TODO (TF2.X)
    #     @interfaces.legacy_get_updates_support
    def get_updates(self, loss, params):
        grads = self.get_gradients(loss, params)
        self.updates = [K.update_add(self.iterations, 1)]

        lr = self.lr

        completed_updates = K.cast(K.tf.floordiv(self.iterations, self.accum_iters), K.floatx())

        if self.initial_decay > 0:
            lr = lr * (1.0 / (1.0 + self.decay * completed_updates))

        t = completed_updates + 1

        lr_t = lr * (K.sqrt(1.0 - K.pow(self.beta_2, t)) / (1.0 - K.pow(self.beta_1, t)))

        # self.iterations incremented after processing a batch
        # batch:              1 2 3 4 5 6 7 8 9
        # self.iterations:    0 1 2 3 4 5 6 7 8
        # update_switch = 1:        x       x    (if accum_iters=4)
        update_switch = K.equal((self.iterations + 1) % self.accum_iters, 0)
        update_switch = K.cast(update_switch, K.floatx())

        ms = [K.zeros(K.int_shape(p), dtype=K.dtype(p)) for p in params]
        vs = [K.zeros(K.int_shape(p), dtype=K.dtype(p)) for p in params]
        gs = [K.zeros(K.int_shape(p), dtype=K.dtype(p)) for p in params]

        if self.amsgrad:
            vhats = [K.zeros(K.int_shape(p), dtype=K.dtype(p)) for p in params]
        else:
            vhats = [K.zeros(1) for _ in params]

        self.weights = [self.iterations] + ms + vs + vhats

        for p, g, m, v, vhat, tg in zip(params, grads, ms, vs, vhats, gs):

            sum_grad = tg + g
            avg_grad = sum_grad / self.accum_iters_float

            m_t = (self.beta_1 * m) + (1.0 - self.beta_1) * avg_grad
            v_t = (self.beta_2 * v) + (1.0 - self.beta_2) * K.square(avg_grad)

            if self.amsgrad:
                vhat_t = K.maximum(vhat, v_t)
                p_t = p - lr_t * m_t / (K.sqrt(vhat_t) + self.epsilon)
                self.updates.append(
                    K.update(vhat, (1 - update_switch) * vhat + update_switch * vhat_t)
                )
            else:
                p_t = p - lr_t * m_t / (K.sqrt(v_t) + self.epsilon)

            self.updates.append(K.update(m, (1 - update_switch) * m + update_switch * m_t))
            self.updates.append(K.update(v, (1 - update_switch) * v + update_switch * v_t))
            self.updates.append(K.update(tg, (1 - update_switch) * sum_grad))
            new_p = p_t

            # Apply constraints.
            if getattr(p, "constraint", None) is not None:
                new_p = p.constraint(new_p)

            self.updates.append(K.update(p, (1 - update_switch) * p + update_switch * new_p))
        return self.updates

    def get_config(self):
        config = {
            "lr": float(K.get_value(self.lr)),
            "beta_1": float(K.get_value(self.beta_1)),
            "beta_2": float(K.get_value(self.beta_2)),
            "decay": float(K.get_value(self.decay)),
            "epsilon": self.epsilon,
            "amsgrad": self.amsgrad,
        }
        base_config = super(AdamAccumulate, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



from keras.models import Model
from keras.utils import plot_model

from medsegpy.modeling.refinenet.refine_module import refine_module

# from resnet import ResNet50


RESNET_INFO = {
    "resnet50": {"layers": [0, 3, 36, 78, 140, -2], "num_filters": [32, 64, 256, 512, 1024, 2048]}
}


def refinenet_model(input_shape=None, backbone="resnet50"):
    raise NotImplementedError("This is not supported on >TF2.0")
    info = RESNET_INFO[backbone]
    downsampling_layers = info["layers"]
    num_filters = info["num_filters"]

    m = ResNet50(include_top=False, weights=None, input_shape=input_shape)  # noqa
    # chop off final ave pooling layer
    m = Model(inputs=m.inputs, outputs=m.layers[downsampling_layers[-1] - 1].output)

    # Add refinenet modules
    assert len(downsampling_layers) == len(num_filters)

    curr_refine = None

    for i in range(len(downsampling_layers))[::-1]:
        x = m.layers[downsampling_layers[i]].output

        x_ins = [x]
        if curr_refine is not None:
            x_ins.append(curr_refine)

        if i == 0:
            num_filters_out = num_filters[i]
        else:
            num_filters_out = num_filters[i - 1]

        curr_refine = refine_module(
            x_ins,
            num_filters_in=num_filters[i],
            num_filters_out=num_filters_out,
            name_prefix="dec_%d" % i,
        )

    return Model(inputs=m.inputs, outputs=curr_refine)


if __name__ == "__main__":
    save_path = "../imgs/refinenet_resnet50.png"
    m = refinenet_model(input_shape=(288, 288, 1))
    plot_model(m, to_file=save_path)



import inspect
from typing import Callable, Sequence, Union

import tensorflow as tf

from medsegpy.utils import env

try:
    import tf.keras.activations as activations
    import tf.keras.backend as K
except ImportError:
    import keras.activations as activations
    import keras.backend as K


def to_numpy(x):
    """Extract numpy equivalent from tensor. Currently does not support TF1.

    Note:
        This method is to simplify data extraction from fixed tensors.
        It should not be used for tensor computations.

    Args:
        x (Tensor): The tensor.

    Returns:
        ndarray: The numpy array.
    """
    if env.is_tf2() and tf.executing_eagerly():
        return x.numpy()
    return K.eval(x)


def get_shape(x):
    """Returns shape of Keras tensor."""
    if hasattr(K, "int_shape"):
        return K.int_shape(x)
    else:
        return K.get_variable_shape(x)


def reduce_tensor(x, reduction="mean", axis=None, weights=None):
    """Reduce tensor along axis.

    If specified, the `weights` tensor will be broadcast when multipied with `x`.
    Weighted mean is computed over `axis` (if specified), else, it is
    computed over the full tensor.

    Args:
        x (Tensor): The tensor.
        reduction (str, optional): Reduction to apply to the tensor.
            One of `"none"` | `"mean"` | `"sum"`.
            * `"none"`: no reduction applied
            * `"mean"`: The weighted mean of the output is taken
            * `"sum"`: The weighted output will be summed
        axis (int(s), optional): The axis to apply reduction to.
        weights (Tensor, optional): A weights tensor broadcastable to shape of x.

    Returns:
        Tensor: Reduced tensor.

    Raises:
        ValueError: If `reduction` not one of `"none"` | `"mean"` | `"sum"`.
    """
    use_weights = weights is not None
    if use_weights:
        x *= weights

    if reduction == "mean" and use_weights:
        ndim = K.ndim(x)
        if axis is None:
            axis = tuple(range(-ndim, 0))
        else:
            axis = _to_negative_index(axis)
        weights_shape = get_shape(weights)
        valid_dims = tuple(
            dim
            for dim in axis
            if abs(dim) <= len(weights_shape) and weights_shape[dim] not in (1, None)
        )
        broadcast_dims = tuple(dim for dim in axis if dim not in valid_dims)
        reduced = K.sum(x, axis=valid_dims, keepdims=True) / K.sum(weights, axis=valid_dims)
        if broadcast_dims:
            reduced = K.mean(reduced, axis=broadcast_dims)
        return reduced
    elif reduction == "mean":
        return K.mean(x, axis=axis)
    elif reduction == "sum":
        return K.sum(x, axis=axis)
    elif reduction in ("none", None):
        return x
    raise ValueError(f"Reduction '{reduction}' unknown.")


def get_activation(act: Union[str, Callable]):
    if isinstance(act, Callable):
        act_name = act.__name__
        act_fn = act
    else:
        act_name = act
        act_fn = activations.__dict__[act]

    args = inspect.signature(act_fn).parameters.keys()
    return act_name, act_fn, args


def _to_negative_index(axis, ndim):
    """Convert positive axes to their negative counterparts."""
    if not isinstance(axis, Sequence):
        axis = (axis,)
    axis = tuple({dim if dim < 0 else -ndim + dim for dim in axis})
    if len(axis) == 1:
        axis = axis[0]
    return axis



import logging
import warnings
from typing import Dict, Sequence, Tuple, Union

import tensorflow as tf
from keras import Input
from keras.engine import get_source_inputs
from keras.layers import BatchNormalization as BatchNorm
from keras.layers import Concatenate, Conv2D, Conv2DTranspose, Dropout, MaxPooling2D

from medsegpy import config
from medsegpy.config import (
    ContextEncoderConfig,
    ContextInpaintingConfig,
    ContextSegmentationConfig,
    ContextUNetConfig,
)
from medsegpy.modeling.layers.convolutional import ConvStandardized2D
from medsegpy.modeling.layers.normalization import GroupNormalization as GroupNorm
from medsegpy.modeling.model_utils import add_sem_seg_activation, zero_pad_like
from medsegpy.modeling.ssl_utils import SelfSupervisedInfo

from ..model import Model
from .build import META_ARCH_REGISTRY, ModelBuilder

logger = logging.getLogger(__name__)


@META_ARCH_REGISTRY.register()
class ContextEncoder(ModelBuilder):
    def __init__(self, cfg: ContextEncoderConfig):
        """
        "Context Encoder" is the name of an encoder network defined in the paper
        "Context Encoders: Feature Learning by Inpainting" (Pathak et al.), which
        learns to create a compact representation of an image that can then be
        used by a decoder network to fill in missing regions of the input image,
        which is called "inpainting".
        For this class, the context encoder architecture will be the same
        as the encoder for a standard 2D UNet. This is different from the original
        paper, but the actual encoder architecture should not matter too much.
        Args:
            cfg: the configuration file for the Context Encoder
        """
        super().__init__(cfg)
        self._cfg = cfg
        self._conv_type = Conv2D
        self._pooler_type = MaxPooling2D
        self._dim = 2
        self._kernel_size = (3, 3)

        # Get type of convolution
        self.weight_standardization = self._cfg.WEIGHT_STANDARDIZATION

        # Get type of layer normalization
        layer_norm_type = self._cfg.NORMALIZATION
        self._layer_norm = None

        if layer_norm_type == "NoNorm":
            self._layer_norm = None
        elif layer_norm_type == "BatchNorm":
            self._layer_norm = BatchNorm
        elif layer_norm_type == "GroupNorm":
            self._layer_norm = GroupNorm
        else:
            warnings.warn(
                f'The normalization type "{layer_norm_type}" '
                f"is not supported. Batch Norm will be used."
            )
            self._layer_norm = BatchNorm
        self._layer_norm_args = self._cfg.NORMALIZATION_ARGS

        # Store number of specific layer types for naming each layer
        self._num_conv = 0
        self._num_bn = 0
        self._num_dropout = 0
        self._num_pool = 0

        # Store pool sizes for decoder network
        self.pool_sizes = []

    def build_model(self, input_tensor=None) -> Model:
        """
        Builds the encoder network and returns the resulting model.
        This implementation will overload the abstract method defined in the
        superclass "ModelBuilder".
        Args:
            input_tensor: The input to the network.
        Returns:
            model: A Model that defines the encoder network.
        """
        # Get architecture hyperparameters
        cfg = self._cfg
        seed = cfg.SEED
        kernel_size = self._kernel_size
        kernel_initializer = {"class_name": cfg.KERNEL_INITIALIZER, "config": {"seed": seed}}
        num_filters = cfg.NUM_FILTERS
        depth = len(num_filters)

        # Build inputs
        if input_tensor is None:
            inputs = self._build_input(cfg.IMG_SIZE)
        else:
            inputs = input_tensor

        # Encoder
        x = inputs
        x_skips = []
        for depth_cnt in range(depth):
            x = self.build_encoder_block(
                x,
                num_filters[depth_cnt],
                kernel_size=kernel_size,
                activation="relu",
                kernel_initializer=kernel_initializer,
                dropout=0.0,
            )

            # Maxpool until penultimate depth.
            if depth_cnt < depth - 1:
                x_skips.append(x)
                pool_size = self._get_pool_size(x)
                self.pool_sizes.append(pool_size)
                self._num_pool += 1
                x = self._pooler_type(pool_size=pool_size, name="encoder_pool_%d" % self._num_pool)(
                    x
                )

        if input_tensor is not None:
            inputs = get_source_inputs(input_tensor)

        model = Model(inputs=inputs, outputs=[x, *x_skips])

        return model

    def build_encoder_block(
        self,
        x: tf.Tensor,
        num_filters: Sequence[int],
        kernel_size: Union[int, Sequence[int]] = 3,
        activation: str = "relu",
        kernel_initializer: Union[str, Dict] = "he_normal",
        dropout: float = 0.0,
    ) -> tf.Tensor:
        """
        Builds one block of the ContextEncoder.
        Each block consists of the following structure:
        [Conv -> Activation] -> BN -> Dropout.
        Args:
            x: Input tensor.
            num_filters: Number of filters to use for each conv layer.
            kernel_size: Kernel size accepted by Keras convolution layers.
            activation: Activation type.
            kernel_initializer: Kernel initializer accepted by
                                `keras.layers.Conv(...)`.
            dropout: Dropout rate.
        Returns:
            Output of encoder block.
        """
        num_conv_layers = len(num_filters)
        for i, filters in enumerate(num_filters):
            self._num_conv += 1
            if self.weight_standardization and i == num_conv_layers - 1:
                x = ConvStandardized2D(
                    filters,
                    kernel_size,
                    padding="same",
                    activation=activation,
                    kernel_initializer=kernel_initializer,
                    name="encoder_conv_%d" % self._num_conv,
                )(x)
            else:
                x = self._conv_type(
                    filters,
                    kernel_size,
                    padding="same",
                    activation=activation,
                    kernel_initializer=kernel_initializer,
                    name="encoder_conv_%d" % self._num_conv,
                )(x)

        if self._layer_norm is not None:
            self._num_bn += 1
            norm_name = "encoder_bn_%d" % self._num_bn
            x = self._layer_norm(**self._layer_norm_args, name=norm_name)(
                x, **SelfSupervisedInfo.set_to_inference(layer_name=norm_name)
            )

        self._num_dropout += 1
        dropout_name = "encoder_dropout_%d" % self._num_dropout
        x = Dropout(rate=dropout, name=dropout_name)(x)

        return x

    @staticmethod
    def _build_input(input_size: Tuple):
        """
        Creates an input tensor of size "input_size".
        Args:
            input_size: The size of the input tensor for the model.
        Returns:
            A symbolic input (i.e. a placeholder) with size "input_size".
        """
        if len(input_size) == 2:
            input_size = input_size + (1,)
        elif len(input_size) != 3:
            raise ValueError("2D ContextEncoder must have an input of shape HxWxC")
        return Input(input_size)

    @staticmethod
    def _get_pool_size(x: tf.Tensor):
        """
        Determines the right size of the pooling filter based on the
        dimensions of the input tensor `x`.
        Args:
            x: The input tensor.
        Returns:
            A list with the same number of elements as dimensions of `x`,
            where each element is either 2 or 3.
        """
        return [2 if d % 2 == 0 or d % 3 != 0 else 3 for d in x.shape.as_list()[1:-1]]


@META_ARCH_REGISTRY.register()
class ContextUNet(ModelBuilder):
    def __init__(self, cfg: ContextUNetConfig):
        """
        This model implements a basic U-Net structure that takes an input
        image, encodes it with an instance of the ContextEncoder model, and
        the decodes the output of the encoder until the decoded output has
        the same resolution as the input image. The model will then end
        there, without any post-processing.
        For this class, the architecture will the same as a standard 2D UNet.
        The encoder will be the ContextEncoder model and the decoder will
        consist of the appropriate type and number of layers to output a
        feature map with the same size as the original input, given the output
        from the encoder model.
        """
        super().__init__(cfg)
        self._cfg = cfg
        self._conv_type = Conv2D
        self._conv_transpose_type = Conv2DTranspose
        self._dim = 2
        self._kernel_size = (3, 3)

        # Get type of convolution
        self.weight_standardization = self._cfg.WEIGHT_STANDARDIZATION

        # Get type of layer normalization
        layer_norm_type = self._cfg.NORMALIZATION
        self._layer_norm = None

        if layer_norm_type == "NoNorm":
            self._layer_norm = None
        elif layer_norm_type == "BatchNorm":
            self._layer_norm = BatchNorm
        elif layer_norm_type == "GroupNorm":
            self._layer_norm = GroupNorm
        else:
            warnings.warn(
                f'The normalization type "{layer_norm_type}" '
                f"is not supported. Batch Norm will be used."
            )
            self._layer_norm = BatchNorm
        self._layer_norm_args = self._cfg.NORMALIZATION_ARGS

        self._kernel_initializer = {
            "class_name": cfg.KERNEL_INITIALIZER,
            "config": {"seed": cfg.SEED},
        }
        self._decoder_suffix = ""

        # Store number of specific layer types for naming each layer
        self._num_conv = 0
        self._num_conv_t = 0
        self._num_norm = 0
        self._num_dropout = 0

    def build_model(self, input_tensor=None) -> Model:
        """
        Builds the full encoder/decoder architecture and returns the
        resulting model.
        This implementation will overload the abstract method defined in the
        superclass "ModelBuilder".
        Args:
            input_tensor: The input to the network.
        Returns:
            model: A Model that defines the full encoder/decoder architecture.
        """
        cfg = self._cfg

        # Get architecture hyperparameters
        num_filters = cfg.NUM_FILTERS
        kernel_size = self._kernel_size

        # Build inputs
        if input_tensor is None:
            inputs = self._build_input(cfg.IMG_SIZE)
        else:
            inputs = input_tensor

        # Read ContextEncoder config
        context_encoder_config = config.get_config("ContextEncoder", create_dirs=False)
        # Overwrite certain config settings
        context_encoder_config.NUM_FILTERS = num_filters
        context_encoder_config.NORMALIZATION = cfg.NORMALIZATION
        context_encoder_config.NORMALIZATION_ARGS = cfg.NORMALIZATION_ARGS
        context_encoder_config.WEIGHT_STANDARDIZATION = cfg.WEIGHT_STANDARDIZATION
        context_encoder_config.SEED = cfg.SEED
        context_encoder_config.KERNEL_INITIALIZER = cfg.KERNEL_INITIALIZER
        context_encoder_config.IMG_SIZE = cfg.IMG_SIZE

        # depth = length of NUM_FILTERS list
        depth = len(num_filters)

        # Build ContextEncoder model
        encoder_name_in_model = "context_encoder_model"
        context_encoder_builder = ContextEncoder(context_encoder_config)
        context_encoder = context_encoder_builder.build_model(inputs)
        context_encoder.name = encoder_name_in_model
        pool_sizes = context_encoder_builder.pool_sizes

        # Pass input through ContextEncoder
        x = inputs

        encoder_outputs = context_encoder(x)
        x = encoder_outputs[0]
        x_skips = encoder_outputs[1:]

        # Decoder
        for i, (x_skip, unpool_size) in enumerate(zip(x_skips[::-1], pool_sizes[::-1])):
            depth_cnt = depth - i - 2
            x = self.build_decoder_block(
                x,
                x_skip,
                num_filters[depth_cnt],
                unpool_size,
                kernel_size=kernel_size,
                activation="relu",
                kernel_initializer=self._kernel_initializer,
                dropout=0.0,
            )

        x = self.post_processing(x)

        if input_tensor is not None:
            inputs = get_source_inputs(input_tensor)

        model = Model(inputs=inputs, outputs=[x])

        return model

    def post_processing(self, x: tf.Tensor):
        """
        Post-processes the output of ContextUnet based on the user's
        preferences.
        The base ContextUnet will not post-process the output. Other models
        that inherit ContextUnet can overload this function to post-process
        the output of the base ContextUnet.
        Args:
            x: Output of base ContextUnet.
        Return:
            Post-processed output.
        """
        return x

    def build_decoder_block(
        self,
        x: tf.Tensor,
        x_skip: tf.Tensor,
        num_filters: Sequence[int],
        unpool_size: Sequence[int],
        kernel_size: Union[int, Sequence[int]] = 3,
        activation: str = "relu",
        kernel_initializer: Union[str, Dict] = "he_normal",
        dropout: float = 0.0,
    ):
        """
        Builds one block of the decoder.
        Each block of the decoder will have the following structure:
        Input -> Transposed Convolution -> Concatenate Skip Connection
        -> Convolution -> BN -> Dropout (optional)
        Args:
            x: Input tensor.
            x_skip: The output of the next highest layer of the
                    ContextEncoder.
            num_filters: Number of filters for both the transposed convolution
                            and regular convolution.
            unpool_size: The stride for the transposed convolutional
                            layer in order to increase the resolution of the
                            input to that of the next highest layer in the
                            ContextEncoder.
            kernel_size: The filter size for all transposed convolutional and
                            regular convolutional layers.
            activation: The activation to use for all transposed convolutional
                            and regular convolution layers.
            kernel_initializer: The kernel initializer for all transposed
                                    convolutional and regular convolutional
                                    layers.
            dropout: Dropout rate.
        Returns:
           Output of the decoder block.
        """
        x_shape = x.shape.as_list()[1:-1]
        x = self._conv_transpose_type(
            num_filters[0],
            kernel_size,
            padding="same",
            strides=unpool_size,
            kernel_initializer=kernel_initializer,
            name=f"conv2d_transpose_{self._num_conv_t}_{self._decoder_suffix}",
        )(x)
        self._num_conv_t += 1
        x_shape = [x * y for x, y in zip(x_shape, unpool_size)]
        x = Concatenate(axis=-1)([zero_pad_like(x, x_skip, x_shape), x_skip])

        if len(num_filters) > 1:
            num_conv_layers = len(num_filters[1:])
            for i, filters in enumerate(num_filters[1:]):
                if self.weight_standardization and i == num_conv_layers - 1:
                    x = ConvStandardized2D(
                        filters,
                        kernel_size,
                        padding="same",
                        activation=activation,
                        kernel_initializer=kernel_initializer,
                        name=f"conv_standardized2d_{self._num_conv}_" f"{self._decoder_suffix}",
                    )(x)
                else:
                    x = self._conv_type(
                        filters,
                        kernel_size,
                        padding="same",
                        activation=activation,
                        kernel_initializer=kernel_initializer,
                        name=f"conv2d_{self._num_conv}_{self._decoder_suffix}",
                    )(x)
                self._num_conv += 1
            if self._layer_norm is not None:
                norm_name = f"normalization_{self._num_norm}_" f"{self._decoder_suffix}"
                x = self._layer_norm(**self._layer_norm_args, name=norm_name)(
                    x, **SelfSupervisedInfo.set_to_inference(layer_name=norm_name)
                )
                self._num_norm += 1
            dropout_name = f"dropout_{self._num_dropout}_{self._decoder_suffix}"
            x = Dropout(rate=dropout, name=dropout_name)(x)
            self._num_dropout += 1
        return x

    @staticmethod
    def _build_input(input_size: Tuple):
        """
        Creates an input tensor of size "input_size".
        Args:
            input_size: The size of the input tensor for the model.
        Returns:
            A symbolic input (i.e. a placeholder) with size "input_size".
        """
        if len(input_size) == 2:
            input_size = input_size + (1,)
        elif len(input_size) != 3:
            raise ValueError("2D ContextInpainting must have an input of shape HxWxC")
        return Input(input_size)


@META_ARCH_REGISTRY.register()
class ContextInpainting(ContextUNet):
    def __init__(self, cfg: ContextInpaintingConfig):
        """
        The ContextInpainting model builds on top of the base ContextUNet model.
        This model receives an input image with some regions cut out and
        filled in with some value, and tries to reconstruct the original
        image given this input. This task was defined in the paper:
        "Context Encoders: Feature Learning by Inpainting" (Pathak et al.).
        The ContextInpainting model will pass the image through the base
        ContextUNet model, and then post-process the output of ContextUNet
        to create a 1-channel output image, which is the model's prediction
        of what it believes to be the original image.
        This model will be used to pre-train the ContextEncoder model, such
        that the compact representations outputted by the ContextEncoder will
        have some useful information about the input dataset. The pretrained
        ContextEncoder model will then be used for the downstream task of
        image segmentation.
        """
        super().__init__(cfg)
        self._cfg = cfg
        self._decoder_suffix = "inpainting_decoder"

    def post_processing(self, x: tf.Tensor):
        # Final convolutional layer to get to cfg.IMG_SIZE[-1] channel output
        x = self._conv_type(
            self._cfg.IMG_SIZE[-1],
            kernel_size=self._kernel_size,
            padding="same",
            kernel_initializer=self._kernel_initializer,
            name="post_process",
        )(x)
        return x


@META_ARCH_REGISTRY.register()
class ContextSegmentation(ContextUNet):
    """
    The ContextSegmentation model will build off the base ContextUNet model,
    just like the ContextInpainting model. However, for this model, the
    output of ContextUNet will be post-processed such that the model
    outputs segmentation probabilities for each segmentation class.
    The ContextSegmentation model will use a ContextEncoder model whose
    weights are pretrained based on the inpainting task.
    For the first pass of training, the ContextEncoder model will be frozen
    and set to inference mode, such that the batchnorm statistics after
    pretraining are not ruined. The rest of the weights for the
    ContextSegmentation model will be trained using
    the training dataset.
    The second pass of training will unfreeze the ContextEncoder model,
    keeping the ContextEncoder in inference mode, and will fine-tune the
    trainable weights of the ContextEncoder model.
    """

    def __init__(self, cfg: ContextSegmentationConfig):
        super().__init__(cfg)
        self._cfg = cfg
        self._decoder_suffix = "segmentation_decoder"

    def post_processing(self, x: tf.Tensor):
        # Returns probabilities for each segmentation class
        cfg = self._cfg
        num_classes = cfg.get_num_classes()
        x = add_sem_seg_activation(
            x,
            num_classes=num_classes,
            conv_type=self._conv_type,
            kernel_initializer=self._kernel_initializer,
            layer_name="post_process",
        )
        return x



import logging

from keras.initializers import glorot_uniform
from keras.layers import Input
from keras.layers.convolutional import Convolution2D

from ..model import Model
from .segnet import SegNet

logger = logging.getLogger(__name__)


class SegNetBottleneck(SegNet):
    """SegNet with single bottleneck (i.e. no pooling at final depth)."""

    def build_model(self):
        input_tensor = self._input_tensor
        input_shape = self._input_shape
        seed = self._seed
        n_labels = self._n_labels
        depth = self._depth
        num_conv_layers = self._num_conv_layers
        num_filters = self._num_filters
        kernel = self._kernel
        pool_size = self._pool_size
        output_mode = self._output_mode
        single_bn = self._single_bn
        conv_act_bn = self._conv_act_bn

        logger.info("Initializing SegNet with seed: {}".format(self._seed))

        inputs = input_tensor if input_tensor else Input(shape=input_shape)

        mask_layers = []

        curr_layer = inputs
        eff_pool_sizes = []

        # Determine pool sizes
        for i in range(depth):
            level = i + 1
            eff_pool_size = pool_size
            divisor = pool_size[0] ** level
            if input_shape[0] % divisor != 0:
                eff_pool_size = (3, 3)
            eff_pool_sizes.append(eff_pool_size)

        # encoder
        logger.info("Building Encoder...")
        for i in range(depth):
            eff_pool_size = eff_pool_sizes[i]
            # Do not pool on bottleneck layer.
            is_bottleneck = i == depth - 1
            n_conv = num_conv_layers[i] - 1 if is_bottleneck else num_conv_layers[i]
            if conv_act_bn:
                curr_layer, l_mask = self._encoder_block_conv_act_bn(
                    curr_layer,
                    level=i + 1,
                    num_conv_layers=n_conv,
                    num_filters=num_filters[i],
                    kernel=kernel,
                    pool_size=eff_pool_size,
                    add_pool=not is_bottleneck,
                )
            else:
                curr_layer, l_mask = self._encoder_block(
                    curr_layer,
                    level=i + 1,
                    num_conv_layers=n_conv,
                    num_filters=num_filters[i],
                    kernel=kernel,
                    pool_size=eff_pool_size,
                    single_bn=single_bn,
                    add_pool=not is_bottleneck,
                )

            mask_layers.append(l_mask)

        # Add additional layer at end:
        if conv_act_bn:
            curr_layer, _ = self._encoder_block_conv_act_bn(
                curr_layer,
                level=depth + 1,
                num_conv_layers=1,
                num_filters=num_filters[depth - 2],
                kernel=kernel,
                pool_size=eff_pool_size,
                add_pool=False,
            )
        else:
            curr_layer, _ = self._encoder_block(
                curr_layer,
                level=depth + 1,
                num_conv_layers=1,
                num_filters=num_filters[depth - 2],
                kernel=kernel,
                pool_size=eff_pool_size,
                single_bn=single_bn,
                add_pool=False,
            )

        logger.info("Building decoder...")
        # decoder
        for i in reversed(range(depth - 1)):
            l_mask = mask_layers[i]
            eff_pool_size = eff_pool_sizes[i]
            if conv_act_bn:
                curr_layer = self._decoder_block_conv_act_bn(
                    curr_layer,
                    l_mask,
                    level=i + 1,
                    num_conv_layers=num_conv_layers[i],
                    num_filters=num_filters[i],
                    num_filters_next=1 if i == 0 else num_filters[i - 1],
                    kernel=kernel,
                    pool_size=eff_pool_size,
                )
            else:
                curr_layer = self._decoder_block(
                    curr_layer,
                    l_mask,
                    level=i + 1,
                    num_conv_layers=num_conv_layers[i],
                    num_filters=num_filters[i],
                    num_filters_next=num_filters[i] if i == 0 else num_filters[i - 1],
                    kernel=kernel,
                    pool_size=eff_pool_size,
                    single_bn=single_bn,
                )

        outputs = Convolution2D(
            n_labels, (1, 1), kernel_initializer=glorot_uniform(seed=seed), activation=output_mode
        )(curr_layer)

        model = Model(inputs=inputs, outputs=outputs, name="segnet_bottleneck")

        return model



"""Metadata catalogs for different datasets.

Metadata stores information like directory paths, mapping from class ids to
name, etc.

Adopted from Facebook's detectron2.
https://github.com/facebookresearch/detectron2
"""
import copy
import logging
import types
from typing import List

from medsegpy.utils.logger import log_first_n

__all__ = ["DatasetCatalog", "MetadataCatalog"]


class DatasetCatalog(object):
    """
    A catalog that stores information about the datasets and how to obtain them.

    It contains a mapping from strings
    (which are names that identify a dataset, e.g. "oai_2d_train")
    to a function which parses the dataset and returns the samples in the
    format of `list[dict]`.

    The returned dicts should be in MedSegPy Dataset format
    (See DATASETS.md for details) if used with the data loader functionalities
    in `data/build.py,data/detection_transform.py`.

    The purpose of having this catalog is to make it easy to choose
    different datasets, by just using the strings in the config.
    """

    _REGISTERED = {}

    @staticmethod
    def register(name, func):
        """
        Args:
            name (str): the name that identifies a dataset,
                e.g. "coco_2014_train".
            func (callable): a callable which takes no arguments and returns a
                list of dicts.
        """
        assert callable(func), "You must register a function with `DatasetCatalog.register`!"
        assert name not in DatasetCatalog._REGISTERED, "Dataset '{}' is already registered!".format(
            name
        )
        DatasetCatalog._REGISTERED[name] = func

    @staticmethod
    def get(name):
        """
        Call the registered function and return its results.

        Args:
            name (str): the name that identifies a dataset,
                e.g. "coco_2014_train".

        Returns:
            list[dict]: dataset annotations.0
        """
        try:
            f = DatasetCatalog._REGISTERED[name]
        except KeyError:
            raise KeyError(
                "Dataset '{}' is not registered! "
                "Available datasets are: {}".format(
                    name, ", ".join(DatasetCatalog._REGISTERED.keys())
                )
            )
        return f()

    @staticmethod
    def list() -> List[str]:
        """
        List all registered datasets.

        Returns:
            list[str]
        """
        return list(DatasetCatalog._REGISTERED.keys())

    @staticmethod
    def clear():
        """
        Remove all registered dataset.
        """
        DatasetCatalog._REGISTERED.clear()


class Metadata(types.SimpleNamespace):
    """
    A class that supports simple attribute setter/getter.
    It is intended for storing metadata of a dataset and make it accessible
    globally.

    Examples:

    .. code-block:: python

        # somewhere when you load the data:
        MetadataCatalog.get("mydataset").thing_classes = ["person", "dog"]

        # somewhere when you print statistics or visualize:
        classes = MetadataCatalog.get("mydataset").thing_classes
    """

    # the name of the dataset
    # set default to N/A so that `self.name` in the errors will not trigger
    # getattr again
    name: str = "N/A"

    _RENAMED = {}

    def __getattr__(self, key):
        if key in self._RENAMED:
            log_first_n(
                logging.WARNING,
                "Metadata '{}' was renamed to '{}'!".format(key, self._RENAMED[key]),
                n=10,
            )
            return getattr(self, self._RENAMED[key])

        raise AttributeError(
            "Attribute '{}' does not exist in the metadata of '{}'. "
            "Available keys are {}.".format(key, self.name, str(self.__dict__.keys()))
        )

    def __setattr__(self, key, val):
        if key in self._RENAMED:
            log_first_n(
                logging.WARNING,
                "Metadata '{}' was renamed to '{}'!".format(key, self._RENAMED[key]),
                n=10,
            )
            setattr(self, self._RENAMED[key], val)

        # Ensure that metadata of the same name stays consistent
        try:
            oldval = getattr(self, key)
            assert oldval == val, (
                "Attribute '{}' in the metadata of '{}' cannot be set "
                "to a different value!\n{} != {}".format(key, self.name, oldval, val)
            )
        except AttributeError:
            super().__setattr__(key, val)

    def as_dict(self):
        """
        Returns all the metadata as a dict.
        Note that modifications to the returned dict will not reflect on the
        Metadata object.
        """
        return copy.copy(self.__dict__)

    def set(self, **kwargs):
        """
        Set multiple metadata with kwargs.
        """
        for k, v in kwargs.items():
            setattr(self, k, v)
        return self

    def get(self, key, default=None):
        """
        Access an attribute and return its value if exists.
        Otherwise return default.
        """
        try:
            return getattr(self, key)
        except AttributeError:
            return default


class MetadataCatalog:
    """
    MetadataCatalog provides access to "Metadata" of a given dataset.

    The metadata associated with a certain name is a singleton: once created,
    the metadata will stay alive and will be returned by future calls to
    `get(name)`.

    It's like global variables, so don't abuse it.
    It's meant for storing knowledge that's constant and shared across the
    execution of the program, e.g.: the class names in OAI iMorphics.
    """

    _NAME_TO_META = {}

    @staticmethod
    def get(name):
        """
        Args:
            name (str): name of a dataset (e.g. oai_2d_train).

        Returns:
            Metadata: The :class:`Metadata` instance associated with this name,
            or create an empty one if none is available.
        """
        assert len(name)
        if name in MetadataCatalog._NAME_TO_META:
            ret = MetadataCatalog._NAME_TO_META[name]
            return ret
        else:
            m = MetadataCatalog._NAME_TO_META[name] = Metadata(name=name)
            return m

    @staticmethod
    def convert_path_to_dataset(path):
        """Convert the dataset path to name for legacy code.

        This method will be phased out in future versions.
        """
        catalog = {
            m.get("scan_root", None): name for name, m in MetadataCatalog._NAME_TO_META.items()
        }
        catalog.pop(None)
        return catalog[path]



import os
import subprocess

from keras.utils import multi_gpu_model

from medsegpy.modeling.model import Model


def get_weights(experiment_dir):
    """Gets the weights file corresponding to lowest validation loss.

    Assumes that only the best weights are stored, so searching for the epoch
    should be enough.
    TODO: remove this assumption.

    Args:
        experiment_dir (str): Experiment directory where weights are stored.

    Returns:
        str: Path to weights h5 file.
    """
    files = os.listdir(experiment_dir)
    max_epoch = -1
    best_file = ""
    for file in files:
        file_fullpath = os.path.join(experiment_dir, file)
        # Ensure the file is an h5 file
        if not (
            os.path.isfile(file_fullpath) and file_fullpath.endswith(".h5") and "weights" in file
        ):
            continue

        # Get file with max epochs
        train_info = file.split(".")[1]
        epoch = int(train_info.split("-")[0])

        if epoch > max_epoch:
            max_epoch = epoch
            best_file = file_fullpath

    if not best_file:
        raise FileNotFoundError("No weights file found in %s" % experiment_dir)

    return best_file


def _check_results_file(base_path):
    """Recursively check for results.txt file."""
    if (base_path is None) or (not os.path.isdir(base_path)) or (base_path == ""):
        return []

    results_filepath = os.path.join(base_path, "results.txt")

    results_paths = []
    if os.path.isfile(results_filepath):
        results_paths.append(results_filepath)

    files = os.listdir(base_path)
    for file in files:
        possible_dir = os.path.join(base_path, file)
        if os.path.isdir(possible_dir):
            subdir_results_files = _check_results_file(possible_dir)
            results_paths.extend(subdir_results_files)

    return results_paths


def get_valid_subdirs(root_dir: str, exist_ok: bool = False):
    """Recursively search for experiments that are ready to be tested.

    Different experiments live in different folders. Based on training protocol,
    we assume that an valid experiment has completed training if its folder
    contains files "config.ini" and "pik_data.dat".

    To avoid recomputing experiments with results, `exist_ok=False` by default.

    Args:
        root_dir (str): Root folder to search.
        exist_ok (:obj:`bool`, optional): If `True`, recompute results for
            experiments.

    Return:
        List[str]: Experiment directories to test.
    """
    no_results = not exist_ok
    if (root_dir is None) or (not os.path.isdir(root_dir)) or (root_dir == []):
        return []

    subdirs = []
    config_path = os.path.join(root_dir, "config.ini")
    pik_data_path = os.path.join(root_dir, "pik_data.dat")
    test_results_dirpath = os.path.join(root_dir, "test_results")
    results_file_exists = len(_check_results_file(test_results_dirpath)) > 0

    # 1. Check if you are a valid subdirectory - must contain a pik data path
    if os.path.isfile(config_path) and os.path.isfile(pik_data_path):
        if (no_results and (not results_file_exists)) or ((not no_results)):
            subdirs.append(root_dir)

    files = os.listdir(root_dir)
    # 2. Recursively search through other subdirectories
    for file in files:
        possible_dir = os.path.join(root_dir, file)
        if os.path.isdir(possible_dir):
            rec_subdirs = get_valid_subdirs(possible_dir, no_results)
            subdirs.extend(rec_subdirs)

    return subdirs


def get_available_gpus(num_gpus: int = None):
    """Get gpu ids for gpus that are >95% free.

    Tensorflow does not support checking free memory on gpus.
    This is a crude method that relies on `nvidia-smi` to
    determine which gpus are occupied and which are free.

    Args:
        num_gpus: Number of requested gpus. If not specified,
            ids of all available gpu(s) are returned.

    Returns:
        List[int]: List of gpu ids that are free. Length
            will equal `num_gpus`, if specified.
    """
    # Built-in tensorflow gpu id.
    assert isinstance(num_gpus, (type(None), int))
    if num_gpus == 0:
        return [-1]

    num_requested_gpus = num_gpus
    num_gpus = (
        len(subprocess.check_output("nvidia-smi --list-gpus", shell=True).decode().split("\n")) - 1
    )

    out_str = subprocess.check_output("nvidia-smi | grep MiB", shell=True).decode()
    mem_str = [x for x in out_str.split() if "MiB" in x]
    # First 2 * num_gpu elements correspond to memory for gpus
    # Order: (occupied-0, total-0, occupied-1, total-1, ...)
    mems = [float(x[:-3]) for x in mem_str]
    gpu_percent_occupied_mem = [
        mems[2 * gpu_id] / mems[2 * gpu_id + 1] for gpu_id in range(num_gpus)
    ]

    available_gpus = [gpu_id for gpu_id, mem in enumerate(gpu_percent_occupied_mem) if mem < 0.05]
    if num_requested_gpus and num_requested_gpus > len(available_gpus):
        raise ValueError(
            "Requested {} gpus, only {} are free".format(num_requested_gpus, len(available_gpus))
        )

    return available_gpus[:num_requested_gpus] if num_requested_gpus else available_gpus


def num_gpus():
    if "CUDA_VISIBLE_DEVICES" not in os.environ or not os.environ["CUDA_VISIBLE_DEVICES"]:
        return 0

    return len(os.environ["CUDA_VISIBLE_DEVICES"].split(","))


class ModelMGPU(Model):
    def __init__(self, ser_model, gpus):
        pmodel = multi_gpu_model(ser_model, gpus)
        self.__dict__.update(pmodel.__dict__)
        self._smodel = ser_model

    def __getattribute__(self, attrname):
        """Override load and save methods to be used from the serial-model. The
        serial-model holds references to the weights in the multi-gpu model.
        """
        # return Model.__getattribute__(self, attrname)
        if "load" in attrname or "save" in attrname:
            return getattr(self._smodel, attrname)

        return super(ModelMGPU, self).__getattribute__(attrname)


class _NoOpScope:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        pass


class NoOpStrategy:
    def scope(self):
        return _NoOpScope()



"""Transforms specific to medical image data.

Many transform operations are typically written with the assumption of 2D data.
However, medical images typically have >2 dimensions.

Some transforms implemented in this file are meant to overload transforms in the
`fvcore.transforms` module.
"""
import inspect
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Sequence, Tuple, TypeVar

import numpy as np

__all__ = [
    "MedTransform",
    "TransformList",
    "CropTransform",
    "ZeroMeanNormalization",
    "FillRegionsWithValue",
    "Swap2DPatches",
]


class MedTransform(ABC):
    """
    Base class for implementations of __deterministic__ transfomations for
    _medical_ image and other data structures. Like the `fvcore.transforms`
    module, there should be a higher-level policy that generates (likely with
    random variations) these transform ops.

    By default, all transforms only handle image and segmentation data types.
    Coordinates, bounding boxes, and polygons are not supported by default.
    However, these methods can be overloaded if generalized methods are
    written for these data types.

    Medical images are seldom in the uint8 format and are not always
    normalized between [0, 1] in the floating point format. Transforms should
    not expect that data is normalized in this format.

    Note, each method may choose to modify the input data in-place for
    efficiency.
    """

    def _set_attributes(self, params: dict = None):
        """
        Set attributes from the input list of parameters.

        Args:
            params (list): list of parameters.
        """

        if params:
            for k, v in params.items():
                if k != "self" and not k.startswith("_"):
                    setattr(self, k, v)

    def state_dict(self, ignore) -> Dict[str, Any]:
        return dict(self.__dict__)

    @abstractmethod
    def apply_image(self, img: np.ndarray):
        """
        Apply the transform on an image.

        Args:
            img (ndarray): of shape NxHxWxC, or HxWxC or HxW. The array can be
                of type uint8 in range [0, 255], or floating point in range
                [0, 1] or [0, 255].
        Returns:
            ndarray: image after apply the transformation.
        """
        pass

    def apply_segmentation(self, segmentation: np.ndarray) -> np.ndarray:
        """
        Apply the transform on a full-image segmentation.
        By default will just perform "apply_image".

        Args:
            segmentation (ndarray): of shape HxW. The array should have integer
            or bool dtype.

        Returns:
            ndarray: segmentation after apply the transformation.
        """
        return self.apply_image(segmentation)

    @classmethod
    def register_type(cls, data_type: str, func: Callable):
        """
        Register the given function as a handler that this transform will use
        for a specific data type.

        Args:
            data_type (str): the name of the data type (e.g., box)
            func (callable): takes a transform and a data, returns the
                transformed data.

        Examples:

        .. code-block:: python

            def func(flip_transform, voxel_data):
                return transformed_voxel_data
            HFlipTransform.register_type("voxel", func)

            # ...
            transform = HFlipTransform(...)
            transform.apply_voxel(voxel_data)  # func will be called
        """
        assert callable(
            func
        ), "You can only register a callable to a MedTransform. " "Got {} instead.".format(func)
        argspec = inspect.getfullargspec(func)
        assert len(argspec.args) == 2, (
            "You can only register a function that takes two positional "
            "arguments to a Transform! Got a function with spec {}".format(str(argspec))
        )
        setattr(cls, "apply_" + data_type, func)


_T = TypeVar("_T")


# pyre-ignore-all-errors
class TransformList:
    """
    Maintain a list of transform operations which will be applied in sequence.
    Attributes:
        transforms (list[Transform])
    """

    def __init__(self, transforms: list):
        """
        Args:
            transforms (list[Transform]): list of transforms to perform.
        """
        super().__init__()
        for t in transforms:
            assert isinstance(t, MedTransform), t
        self.transforms = transforms

    def _apply(self, x: _T, meth: str) -> _T:
        """
        Apply the transforms on the input.
        Args:
            x: input to apply the transform operations.
            meth (str): meth.
        Returns:
            x: after apply the transformation.
        """
        for t in self.transforms:
            x = getattr(t, meth)(x)
        return x

    def __getattr__(self, name: str):
        """
        Args:
            name (str): name of the attribute.
        """
        if name.startswith("apply_"):
            return lambda x: self._apply(x, name)
        raise AttributeError("TransformList object has no attribute {}".format(name))

    def __add__(self, other: "TransformList") -> "TransformList":
        """
        Args:
            other (TransformList): transformation to add.
        Returns:
            TransformList: list of transforms.
        """
        others = other.transforms if isinstance(other, TransformList) else [other]
        return TransformList(self.transforms + others)

    def __iadd__(self, other: "TransformList") -> "TransformList":
        """
        Args:
            other (TransformList): transformation to add.
        Returns:
            TransformList: list of transforms.
        """
        others = other.transforms if isinstance(other, TransformList) else [other]
        self.transforms.extend(others)
        return self

    def __radd__(self, other: "TransformList") -> "TransformList":
        """
        Args:
            other (TransformList): transformation to add.
        Returns:
            TransformList: list of transforms.
        """
        others = other.transforms if isinstance(other, TransformList) else [other]
        return TransformList(others + self.transforms)

    def __len__(self) -> int:
        """
        Returns:
            Number of transforms contained in the TransformList.
        """
        return len(self.transforms)


class CropTransform(MedTransform):
    """Crop nD volumes.

    Volumes will be cropped as ``vol[..., z:z+d, y:y+h, x:x+w]``.

    Args:
        coords0 (:obj:`int(s)`): Initial coordinates. In order x,y,z,...
        crop_size (:obj:`int(s)`): Crop size. In order w,h,d,...
    """

    def __init__(self, coords0: Sequence[int], crop_size: Sequence[int]):
        assert len(coords0) == len(crop_size)
        super().__init__()
        window = [slice(c, c + s) for c, s in zip(coords0, crop_size)] + [slice()]
        window.insert(0, Ellipsis)
        self._set_attributes({"window": window})

    def apply_image(self, img: np.ndarray) -> np.ndarray:
        """Crop the image(s).

        Args:
            img (ndarray): of shape >=`len(self.window)`

        Returns:
            ndarray: cropped image(s).
        """

        return img[self.window]


class ZeroMeanNormalization(MedTransform):
    """Zero mean unit variance normalization.

    Volumes are dynamically normalized to be zero-mean and unit variance.

    Args:
        axis (int|Tuple[int]): Axis or axes along which the statistics
            (means and standard deviations) are computed. The default is
            to compute the statistics over the spatial dimensions only
            (i.e. batch and channel dims are ignored).
    """

    def __init__(self, axis=None):
        super().__init__()
        self._set_attributes({"axis": axis})

    def apply_image(self, img: np.ndarray):
        """Crop the image(s).

        Args:
            img (ndarray): of shape `CxHxWx...`

        Returns:
            ndarray: zero-mean, unit variance image.
        """
        axis = self.axis
        if axis is None:
            axis = tuple(range(1, img.ndim))
        mean = img.mean(axis=axis, keepdims=True)
        std = img.mean(axis=axis, keepdims=True)
        return (img - mean) / std

    def apply_segmentation(self, segmentation: np.ndarray):
        """Segmentation should not be normalized."""
        return segmentation


class AffineNormalization(MedTransform):
    """Affine normalization by ``scale`` and ``bias``.

    Image ``x`` will be normalized as ``x = (x - bias) / scale``.
    If ``bias`` or ``scale`` are 1D vectors, they will be applied along
    the channel dimension.

    Args:
        scale (float | array-like): Scaling parameter.
            If this is 1D, it will be applied along the channel dimension.
        bias (float | array-like): Bias parameter.
            If this is 1D, it will be applied along the channel dimension.
    """

    def __init__(self, scale=1.0, bias=0.0):
        super().__init__()
        self._set_attributes({"scale": scale, "bias": bias})

    def _broadcast_param(self, param, ndim):
        param = np.asarray(param)
        if param.ndim == 1:
            param = param.reshape((param.shape[0],) + (1,) * (ndim - 2))
        return param

    def apply_image(self, img: np.ndarray):
        """Crop the image(s).

        Args:
            img (ndarray): of shape `NxCx...`

        Returns:
            ndarray: zero-mean, unit variance image.
        """
        scale = self._broadcast_param(self.scale, img.ndim)
        bias = self._broadcast_param(self.bias, img.ndim)
        return (img - bias) / scale

    def apply_segmentation(self, segmentation: np.ndarray):
        """Segmentation should not be normalized."""
        return segmentation


class FlipTransform(MedTransform):
    """Flip image along axis or axes.

    Input images and segmentations will be flipped along ``axis``.
    In most cases, these axes should correspond to spatial dimensions.
    If a channel dimension is specified, the segmentation channels will
    also flip.

    Args:
        axis (int|Tuple[int]): Axis or axes along which to flip.
            Defaults to no flipping.
    """

    def __init__(self, axis=None) -> None:
        super().__init__()
        self._set_attributes({"axis": axis})

    def apply_image(self, img: np.ndarray) -> np.ndarray:
        """Crop the image(s).

        Args:
            img (ndarray): of shape >=`len(self.window)`

        Returns:
            ndarray: cropped image(s).
        """
        axis = self.axis
        if not axis:
            return img
        return np.flip(img, axis=self.axis)


class Windowing(MedTransform):
    """Clip image magnitude between values `lower`, `upper`.

    If multiple lower/upper bound pairs are provided, the output will be stacked
    along the last dimension.
    """

    def __init__(self, bounds: Sequence[Tuple[int, int]]):
        """
        Args:
            bounds: `[(10,20)]` or `[(10,20), (40, 100)]` etc.
        """
        self._set_attributes(locals())
        super().__init__()

    def apply_image(self, img: np.ndarray):
        """Crop the image(s).

        Args:
            img (ndarray): of shape NxCxDxHxW, or NxDxHxW or DxHxW.

        Returns:
            ndarray: zero-mean, unit variance image.
        """
        imgs = []
        bounds = self.bounds
        for lower, upper in bounds:
            imgs.append(np.clip(img, a_min=lower, a_max=upper))

        if len(imgs) == 1:
            return imgs[0]
        elif img.shape[-1] == 1:
            return np.concatenate(imgs, axis=-1)
        else:
            return np.stack(imgs, axis=-1)

    def apply_segmentation(self, segmentation: np.ndarray):
        """Segmentation should not be windowed."""
        return segmentation


class FillRegionsWithValue(MedTransform):
    """Fills rectangular regions of an image with a constant value.

    This transform was included to help with the implementation of Coarse
    Dropout, located in "medsegpy/data/transforms/transform_gen.py".

    If a stack of images is provided as input to this transform, the
    transform will be applied independently for each image. As a result,
    the regions chosen for one image may be different from the regions
    chosen for another image.

    The implementation is adapted from a similar function in:
    https://github.com/albumentations-team/albumentations/blob/master/albumentations/augmentations/functional.py.
    """

    def __init__(self, hole_mask: np.ndarray, fill_value: float = 0.0):
        """
        Args:
            hole_mask: A mask, the same size as the image, indicating which
                pixels should be filled with `fill_value`.
            fill_value: The value used to fill in each hole.
        """
        self._set_attributes(locals())
        super().__init__()

    def apply_image(self, img: np.ndarray):
        """Applies the transform to an image array, which may consist of a
        stack of images.

        If the images have > 1 channel (the last dimension of the input
        image array is > 1), the same holes will be filled in for
        all channels.

        Args:
            img: A N x H x W x C array, containing N images of height H,
                    width W, and consisting of C channels.

        Returns:
            img_filled: A N x H x W x C array, containing the transformed
                image.
        """
        assert img.shape == self.hole_mask.shape, (
            f"Shape of 'hole_mask' ({self.hole_mask.shape}) must match shape "
            f"of 'image' ({img.shape})"
        )
        img_filled = img.copy()
        img_filled[np.array(self.hole_mask, dtype=bool)] = self.fill_value
        return img_filled

    def apply_segmentation(self, segmentation: np.ndarray) -> np.ndarray:
        """Segmentation should not be modified."""
        return segmentation


class Swap2DPatches(MedTransform):
    """Swaps pairs of patches in an image."""

    def __init__(self, patch_pairs):
        """
        Args:
            patch_pairs: A list of dictionaries, where each dictionary has
                            the following structure:

                            {"patch_1": (x1, y1, x2, y2),
                             "patch_2": (x1, y1, x2, y2)}

                            (x1, y1) is the coordinate of the top-left
                            corner of the patch, and (x2, y2) is the
                            coordinate of the bottom-right corner of the
                            patch.
        """
        super().__init__()
        self._set_attributes(locals())

    def apply_image(self, img: np.ndarray):
        """Applies the transform to an image array, which may consist of a
        stack of images.

        If the input image array consists of a stack of images (i.e. the
        first dimension is > 1), the appropriate pair of patches will be
        swapped based on the index of the image in the input image array.
        For example, the patches to be swapped for the first image
        (index 0 in the first dimension of the input image array) are
        located index 0 of "patch_pairs".

        If the images have > 1 channel (the last dimension of the input
        image array is > 1), the same pairs of patches will be swapped
        for all channels.

        Args:
            img: A N x H x W x C array, containing N images of height H,
                    width W, and consisting of C channels.

        Returns:
            img_modified: A N x H x W x C array containing the modified
                            image.
        """
        assert img.shape[0] == len(
            self.patch_pairs
        ), "Number of images does not equal the number of patch pairs!"
        img_modified = img.copy()
        for i, patch_pair in enumerate(self.patch_pairs):
            tl = patch_pair[0]
            br = patch_pair[1]
            for num_pair in range(tl.shape[-1]):
                # Get coordinates of both patches
                x1_p1, y1_p1 = tl[:, 0, num_pair]
                x2_p1, y2_p1 = br[:, 0, num_pair]
                x1_p2, y1_p2 = tl[:, 1, num_pair]
                x2_p2, y2_p2 = br[:, 1, num_pair]

                # Swap patches
                tmp_img_patch = np.copy(img_modified[i, y1_p1:y2_p1, x1_p1:x2_p1, :])
                img_modified[i, y1_p1:y2_p1, x1_p1:x2_p1, :] = np.copy(
                    img_modified[i, y1_p2:y2_p2, x1_p2:x2_p2, :]
                )
                img_modified[i, y1_p2:y2_p2, x1_p2:x2_p2, :] = tmp_img_patch

        return img_modified

    def apply_segmentation(self, segmentation: np.ndarray) -> np.ndarray:
        """Segmentation should not be modified."""
        return segmentation


class NoOpTransform(MedTransform):
    """
    A transform that does nothing.
    """

    def __init__(self):
        super().__init__()

    def apply_image(self, img: np.ndarray) -> np.ndarray:
        return img

    def inverse(self) -> MedTransform:
        return self

    def __getattr__(self, name: str):
        if name.startswith("apply_"):
            return lambda x: x
        raise AttributeError("NoOpTransform object has no attribute {}".format(name))



"""Metrics Processor.

A processor keeps track of task-specific metrics for different classes.
It should not be used to keep track of non-class-specific metrics, such as
runtime.
"""
import inspect
import logging
from abc import ABC, abstractmethod
from collections import OrderedDict
from enum import Enum
from typing import Callable, Collection, Dict, Sequence, Union

import numpy as np
import pandas as pd
import tabulate
from medpy.metric import assd

logger = logging.getLogger(__name__)

__all__ = ["Metric", "DSC", "VOE", "CV", "ASSD", "Precision", "Recall", "MetricsManager"]


def flatten_non_category_dims(xs: Sequence[np.ndarray], category_dim: int = None):
    """Flattens all non-category dimensions into a single dimension.

    Args:
        xs (ndarrays): Sequence of ndarrays with the same category dimension.
        category_dim: The dimension/axis corresponding to different categories.
            i.e. `C`. If `None`, behaves like `np.flatten(x)`.

    Returns:
        ndarray: Shape (C, -1) if `category_dim` specified else shape (-1,)
    """
    single_input = isinstance(xs, np.ndarray)
    if single_input:
        xs = [xs]
    if category_dim is not None:
        dims = (xs[0].shape[category_dim], -1)
        xs = (np.moveaxis(x, category_dim, 0).reshape(dims) for x in xs)
    else:
        xs = (x.flatten() for x in xs)

    if single_input:
        xs = list(xs)[0]

    return xs


def rms_cv(y_pred: np.ndarray, y_true: np.ndarray, dim=None):
    """Compute root-mean-squared coefficient of variation.

    This is typically done to compare intra-method variability.
    For example if multiple measurements are taken using the same method.
    However, in many segmentation manuscripts, this is equation is also
    used.

    This quantity is symmetric.

    Args:
        y_pred (ndarray): Measurements from trial 1.
        y_true (ndarray): Measurements from trial 2.
        dim (int, optional): Dimension/axis over which to compute metric.
            If `None`, all dimensions will be reduced.

    Returns:
        ndarray: If `dim=None`, scalar value.
    """
    stds = np.std([y_pred, y_true], axis=0)
    means = np.mean([y_pred, y_true], axis=0)
    cv = stds / means
    return np.sqrt(np.mean(cv**2, axis=dim))


def rmse_cv(y_pred: np.ndarray, y_true: np.ndarray, dim=None):
    """Compute root-mean-squared error coefficient of variation.

    This quantity is not symmetric.

    Args:
        y_pred (ndarray): Predicted measurements.
        y_true (ndarray): Ground-truth/baseline measurements.
        dim (int, optional): Dimension/axis over which to compute metric.
            If `None`, all dimensions will be reduced.

    Returns:
        ndarray: If `dim=None`, scalar value.
    """
    rmse = np.sqrt(np.mean((y_pred - y_true) ** 2, axis=dim))
    means = np.absolute(np.mean(y_true, axis=dim))
    return rmse / means


class Reductions(Enum):
    RMS_CV = 1, "RMS-CV", rms_cv
    RMSE_CV = 2, "RMSE-CV", rmse_cv

    def __new__(cls, value: int, display_name: str, func: Callable):
        """
        Args:
            value (int): Unique integer value.
            patterns (`List[str]`): List of regex patterns that would match the
                hostname on the compute cluster. There can be multiple hostnames
                per compute cluster because of the different nodes.
            save_dir (str): Directory to save data to.
        """
        obj = object.__new__(cls)
        obj._value_ = value

        obj.display_name = display_name
        obj.func = func

        return obj


class Metric(Callable, ABC):
    """Interface for new metrics.

    A metric should be implemented as a callable with explicitly defined
    arguments. In other words, metrics should not have `**kwargs` or `**args`
    options in the `__call__` method.

    While not explicitly constrained to the return type, metrics typically
    return float value(s). The number of values returned corresponds to the
    number of categories.

    * metrics should have different name() for different functionality.
    * `category_dim` duck type if metric can process multiple categories at
        once.

    To compute metrics:

    .. code-block:: python

        metric = Metric()
        results = metric(...)
    """

    def __init__(self, units: str = ""):
        self.units = units

    def name(self):
        return type(self).__name__

    def display_name(self):
        """Name to use for pretty printing and display purposes."""
        name = self.name()
        return "{} {}".format(name, self.units) if self.units else name

    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass


class DSC(Metric):
    """Dice score coefficient."""

    def __call__(self, y_pred, y_true, category_dim: int = None):
        y_pred = y_pred.astype(np.bool)
        y_true = y_true.astype(np.bool)
        y_pred, y_true = flatten_non_category_dims((y_pred, y_true), category_dim)

        size_i1 = np.count_nonzero(y_pred, -1)
        size_i2 = np.count_nonzero(y_true, -1)
        intersection = np.count_nonzero(y_pred & y_true, -1)
        smooth = np.finfo(float).eps

        return (2.0 * intersection + smooth) / (size_i1 + size_i2 + smooth)


class VOE(Metric):
    """Volumetric overlap error."""

    def __call__(self, y_pred, y_true, category_dim: int = None):
        y_pred = y_pred.astype(np.bool)
        y_true = y_true.astype(np.bool)
        y_pred, y_true = flatten_non_category_dims((y_pred, y_true), category_dim)

        intersection = np.count_nonzero(y_true & y_pred, axis=-1)
        union = np.count_nonzero(y_true | y_pred, axis=-1).astype(np.float)

        return 1 - intersection / union


class CV(Metric):
    """Coefficient of variation."""

    def __call__(self, y_pred, y_true, category_dim: int = None):
        y_pred = y_pred.astype(np.bool)
        y_true = y_true.astype(np.bool)
        y_pred, y_true = flatten_non_category_dims((y_pred, y_true), category_dim)

        size_i1 = np.count_nonzero(y_pred, -1)
        size_i2 = np.count_nonzero(y_true, -1)

        std = np.std([size_i1, size_i2], axis=0)
        mean = np.mean([size_i1, size_i2], axis=0)

        return std / mean


class ASSD(Metric):
    """Average symmetric surface distance."""

    def __call__(self, y_pred, y_true, spacing=None, connectivity=1):
        # TODO: check why this statement is needed.
        if not connectivity:
            connectivity = 1
        return assd(y_pred, y_true, voxelspacing=spacing, connectivity=connectivity)


class Precision(Metric):
    def __call__(self, y_pred, y_true, category_dim: int = None):
        y_pred = y_pred.astype(np.bool)
        y_true = y_true.astype(np.bool)
        y_pred, y_true = flatten_non_category_dims((y_pred, y_true), category_dim)

        tp = np.count_nonzero(y_pred & y_true, -1)
        fp = np.count_nonzero(y_pred & ~y_true, -1)

        return tp / (tp + fp)


class Recall(Metric):
    def __call__(self, y_pred, y_true, category_dim: int = None):
        y_pred = y_pred.astype(np.bool)
        y_true = y_true.astype(np.bool)
        y_pred, y_true = flatten_non_category_dims((y_pred, y_true), category_dim)

        tp = np.count_nonzero(y_pred & y_true, -1)
        fn = np.count_nonzero(~y_pred & y_true, -1)

        return tp / (tp + fn)


class L2NORM(Metric):
    def __call__(self, y_pred, y_true):
        assert len(y_pred.shape) == 4 or len(y_pred.shape) == 3, "y_pred must be 3D or 4D"
        if len(y_pred.shape) == 3:
            y_pred = np.expand_dims(y_pred, axis=-1)
            y_true = np.expand_dims(y_true, axis=-1)

        # Reshape to N x H x W x C
        y_pred = np.transpose(y_pred, (2, 0, 1, 3))
        y_true = np.transpose(y_true, (2, 0, 1, 3))

        # Calculate L2 Norm per slice
        # -- Shape: (N x C)
        l2_norm_per_slice = np.sqrt(np.sum(np.square(y_pred - y_true), axis=(1, 2), keepdims=True))
        val_per_slice = np.squeeze(l2_norm_per_slice, axis=(1, 2))

        # Calculate L2 Norm per category
        # -- Shape: (C,)
        avg_l2_norm = np.mean(val_per_slice, axis=0)
        val_per_category = np.transpose(avg_l2_norm)

        return val_per_category


_BUILT_IN_METRICS = {x.__name__: x for x in Metric.__subclasses__()}


class MetricsManager:
    """A class to manage and compute metrics.

    Metrics will be calculated for the categories specified during
    instantiation. All metrics are assumed to be calculated for those
    categories.

    Metrics are indexed by their string representation as returned by `name()`.
    They are also computed in the order they were added.

    To compute metrics, use this class as a callable. See `__call__` for more
    details.

    Attributes:
        class_names (Sequence[str]): Category names (in order).

    To calculate metrics:

        .. code-block:: python

        manager = MetricsManager(
            category_names=("tumor", "no tumor")
            metrics=(DSC(), VOE())
        )

        for scan_id, x, y_pred, y_true in zip(ids, xs, preds, ground_truths):
            # Compute metrics per scan.
            manager(scan_id, x=x, y_pred=y_pred, y_true=y_true)

    To get number of scans that have been processed:

        .. code-block:: python

        num_scans = len(manager)
    """

    def __init__(self, class_names: Collection[str], metrics: Sequence[Union[Metric, str]] = None):
        self._scan_data = OrderedDict()
        self.class_names = class_names
        self._metrics: Dict[str, Metric] = OrderedDict()
        self._metric_pairs: Dict[(str, str), Sequence[Reductions]] = OrderedDict()  # noqa
        if metrics:
            self.add_metrics(metrics)
        self.category_dim = -1 if len(class_names) > 1 else None
        self.runtimes = []
        self._is_data_stale = False

    def metrics(self):
        """Returns names of current metrics."""
        return self._metrics.keys()

    def add_metrics(self, metrics: Sequence[Union[Metric, str]]):
        """Add metrics to compute.

        Metrics with the same `name()` cannot be added.

        Args:
            metrics (Metric(s)/str(s)): Metrics to compute. `str` values should
                only be used for built-in metrics.

        Raises:
            ValueError: If `metric.name()` already exists. Metrics with the
                same name mean the same computation will be done twice, which
                is not a supported feature.
        """
        if isinstance(metrics, (Metric, str)):
            metrics = [metrics]
        for m in metrics:
            if isinstance(m, str):
                m = _BUILT_IN_METRICS[m]()
            m_name = m.name()
            if m_name in self._metrics:
                raise ValueError("Metric {} already exists".format(m_name))
            self._metrics[m_name] = m

    def register_pairs(
        self,
        pred_metric: str,
        base_metric: str,
        reductions: Union[Reductions, Sequence[Reductions]],
        name: str = None,
    ):
        if pred_metric not in self._metrics:
            raise ValueError("`pred_metric` '{}' not found".format(pred_metric))
        if base_metric not in self._metrics:
            raise ValueError("`base_metric` '{}' not found".format(base_metric))

        key = (pred_metric, base_metric)
        if key in self._metric_pairs:
            raise ValueError("Pair {} already registered".format(key))

        if name is None:
            name = self._metrics[pred_metric].__class__.__name__
        self._metric_pairs[key] = (reductions, name)

    def remove_metrics(self, metrics: Union[str, Sequence[str]]):
        """Remove metrics to compute.

        Args:
            metrics (`str(s)`): Names of metrics to remove.
        """
        if isinstance(metrics, str):
            metrics = [metrics]
        pairs_to_remove = []
        for m in metrics:
            del self._metrics[m]
            pairs_to_remove.extend([x for x in self._metric_pairs if m in x])

        for pair in pairs_to_remove:
            del self._metric_pairs[pair]

    def __call__(
        self,
        scan_id: str,
        x: np.ndarray = None,
        y_pred: np.ndarray = None,
        y_true: np.ndarray = None,
        runtime: float = np.nan,
        **kwargs,
    ) -> str:
        """Compute metrics for a scan.

        Args:
            scan_id (str): The scan/example identifier
            x (ndarray, optional): The input `x` accepted by most metrics.
            y_pred (ndarray, optional): The predicted output.
                For most metrics, should be binarized.
                If computing for multiple classes, last dimension should
                index different categories in the order of `self.class_names`.
            y_true (ndarray, optional): The binarized ground truth output.
                For multiple classes, format like `y_pred`.
            runtime (float, optional): The compute time. If specified, logged
                as an additional metric.

        Returns:
            str: A summary of the results for the scan.
        """
        # Hacky way to define some expected values but still add them
        # to kwargs for future processing.
        for k, v in locals().items():
            if k in ["x", "y_pred", "y_true"] and v is not None:
                kwargs[k] = v

        num_classes = len(self.class_names)
        metrics_names = [m.display_name() for m in self._metrics.values()]
        metrics_data = []
        for metric in self._metrics.values():
            args = inspect.getfullargspec(metric).args

            params = {name: kwargs.get(name) for name in args if name not in ["self"]}

            if num_classes > 1 and "category_dim" in args:
                params["category_dim"] = self.category_dim
                metrics_data.append(metric(**params))
            else:
                metrics_data.append(
                    [
                        metric(
                            **{
                                k: v[..., c] if k in ["y_pred", "y_true", "x"] else v
                                for k, v in params.items()
                            }
                        )
                        for c in range(num_classes)
                    ]
                )

        metrics_data = pd.DataFrame(metrics_data, index=metrics_names, columns=self.class_names)
        metrics_data.replace([np.inf, -np.inf], np.nan)
        self.runtimes.append(runtime)

        strs = [
            "{}: {:0.3f}".format(n, v)
            for n, v in zip(metrics_names, np.nanmean(metrics_data, axis=1))
        ]

        self._scan_data[scan_id] = metrics_data
        self._is_data_stale = True

        return ", ".join(strs)

    def _compute_metric_pairs(self, arr):
        pair_names = []
        vals = []
        names = list(self._metrics.keys())
        for (pred_key, base_key), (reds, name) in self._metric_pairs.items():
            pred_idx, base_idx = names.index(pred_key), names.index(base_key)
            preds = arr[:, pred_idx, ...]
            base = arr[:, base_idx, ...]

            vals.extend([r.func(preds, base, dim=0) for r in reds])
            e_names = ["{} {}".format(r.display_name, name) for r in reds]
            pair_names.extend(e_names)

        vals = np.stack(vals, axis=0)
        df = pd.DataFrame(vals, index=pair_names, columns=self.class_names)
        return df

    def scan_summary(self, scan_id, delimiter: str = ", ") -> str:
        """Get summary of results for a scan.

        Args:
            scan_id: Scan id for which to summarize results.
            delimiter (`str`, optional): Delimiter between different metrics.

        Returns:
            str: A summary of metrics for the scan. Values are averaged across
                all categories.
        """
        scan_data = self._scan_data[scan_id]
        avg_data = scan_data.mean(axis=1)

        strs = ["{}: {:0.3f}".format(n, avg_data[n]) for n in avg_data.index.tolist()]

        return delimiter.join(strs)

    def data_frame(self):
        raw_arr = np.stack([np.asarray(x) for x in self._scan_data.values()], axis=0)
        arr = np.nanmean(raw_arr, axis=0)
        names = [m.display_name() for m in self._metrics.values()]

        df = pd.DataFrame(arr, index=names, columns=self.class_names)
        if self._metric_pairs:
            df_pairs = self._compute_metric_pairs(raw_arr)
            df = pd.concat([df, df_pairs])

        return df

    def summary(self):
        """Get summary of results overall scans.

        Returns:
            str: Tabulated summary. Rows=metrics. Columns=classes.
        """
        df = self.data_frame()

        return tabulate.tabulate(df, headers=self.class_names) + "\n"

    def data(self):
        """
        TODO: Determine format
        """
        df = pd.concat(self._scan_data.values(), keys=self._scan_data.keys())
        return {"runtimes": self.runtimes, "scan_data": df}

    def __len__(self):
        return len(self._scan_data)

    def __getitem__(self, scan_id):
        return self._scan_data[scan_id]



import os
import re
import socket
import warnings
from typing import Sequence, Union

import yaml
from fvcore.common.file_io import PathManager

from medsegpy.utils.env import settings_dir

# Path to the repository directory.
_REPO_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "../.."))


class Cluster:
    """Tracks config of different nodes/clusters.

    This class is helpful for managing different storage paths across different
    nodes/clusters without the overhead of duplicating the codebase across
    multiple nodes.

    To identify the current node, we inspect the hostname.
    This can be problematic if two machines have the same hostname, though
    this has not been an issue as of yet.

    DO NOT use the node's public ip address to identify it. Not only is this not
    returned by ``socket.hostname()``, but there are also some security issues.

    Note:
        This class is not thread safe. Saving/deleting configs should be done on
        the main thread.
    """

    def __init__(
        self,
        name: str,
        patterns: Union[str, Sequence[str]],
        data_dir: str = None,
        results_dir: str = None,
    ):
        """
        Args:
            name (str): The name of the cluster. Name is case-sensitive.
            patterns (Sequence[str]): Regex pattern(s) for identifying cluster.
                Cluster will be identified by
                ``any(re.match(p, socket.gethostname()) for p in patterns)``.
            data_dir (str, optional): The data directory. Defaults to
                ``os.environ['MEDSEGPY_RESULTS']`` or ``"./datasets"``.
            results_dir (str, optional): The results directory. Defaults to
                `"os.environ['MEDSEGPY_DATASETS']"` or ``"./results"``.
        """
        self.name = name

        if isinstance(patterns, str):
            patterns = patterns
        self.patterns = patterns

        self._data_dir = data_dir
        self._results_dir = results_dir

    @property
    def data_dir(self):
        path = self._data_dir
        if not path:
            path = os.environ.get("MEDSEGPY_DATASETS", "./datasets")
        return PathManager.get_local_path(path)

    @property
    def results_dir(self):
        path = self._results_dir
        if not path:
            path = os.environ.get("MEDSEGPY_RESULTS", "./results")
        return PathManager.get_local_path(path)

    def save(self):
        """Save cluster config to yaml file."""
        filepath = self.filepath()
        os.makedirs(os.path.dirname(filepath), exist_ok=True)
        with open(filepath, "w") as f:
            data = {(k[1:] if k.startswith("_") else k): v for k, v in self.__dict__.items()}
            yaml.safe_dump(data, f)

    def delete(self):
        """Deletes the config file for this cluster."""
        filepath = self.filepath()
        if os.path.isfile(filepath):
            os.remove(filepath)

    def filepath(self):
        """Returns config file path.

        Note:
            This does not guarantee the config exists. To save the cluster config to a file,
            use `save()`.

        Returns:
            str: The config file path.
        """
        return os.path.join(self._config_dir(), f"{self.name}.yaml")

    @property
    def save_dir(self):
        """Deprecated: Legacy alias for `self.results_dir`"""
        warnings.warn(
            "`save_dir` is deprecated and will be removed in v0.0.2. Use `results_dir` instead",
            FutureWarning,
        )
        return self.results_dir

    @classmethod
    def all_clusters(cls):
        config_dir = cls._config_dir()
        clusters = []
        if os.path.isdir(config_dir):
            files = sorted(os.listdir(config_dir))
            for f in files:
                clusters.append(cls.from_config(os.path.join(config_dir, f)))
        return clusters

    @classmethod
    def cluster(cls):
        """Searches saved clusters by regex matching with hostname.

        Note:
            The cluster must have been saved to a config file. Also, if
            there are multiple cluster matches, only the first (sorted alphabetically)
            will be returned.

        Returns:
            Cluster: The current cluster.
        """
        clusters = cls.all_clusters()
        hostname = socket.gethostname()
        for clus in clusters:
            if any(re.match(p, hostname) for p in clus.patterns):
                return clus
        return _UNKNOWN

    @classmethod
    def from_config(cls, name):
        """
        Args:
            name (str): Cluster name or path to config file.
        Returns:
            Cluster: The Cluster object
        """
        if not os.path.isfile(name):
            filepath = os.path.join(cls._config_dir(), f"{name}.yaml")
        else:
            filepath = name
        if not os.path.isfile(filepath):
            raise FileNotFoundError(f"Config file not found: {name}")

        with open(filepath, "r") as f:
            cfg = yaml.safe_load(f)
        return cls(**cfg)

    @staticmethod
    def _config_dir():
        return os.path.join(settings_dir(), "clusters")

    @staticmethod
    def working_cluster() -> "Cluster":
        return _CLUSTER

    @staticmethod
    def set_working_cluster(cluster=None):
        """Sets the working cluster.

        Args:
            cluster (`str` or `Cluster`): The cluster name or cluster.
                If ``None``, will reset cluster to _UNKNOWN, meaning default
                data and results dirs will be used.
        """
        set_cluster(cluster)

    def __repr__(self):
        return "Cluster({})".format(
            ", ".join("{}={}".format(k, v) for k, v in self.__dict__.items())
        )


def set_cluster(cluster: Union[str, Cluster] = None):
    """Sets the working cluster.

    Args:
        cluster (`str` or `Cluster`): The cluster name or cluster.
            If ``None``, will reset cluster to _UNKNOWN, meaning default
            data and results dirs will be used.
    """
    if cluster is None:
        cluster = _UNKNOWN
    elif isinstance(cluster, str):
        if cluster.lower() == _UNKNOWN.name.lower():
            cluster = _UNKNOWN
        else:
            cluster = Cluster.from_config(cluster)
    global _CLUSTER
    _CLUSTER = cluster


_UNKNOWN = Cluster("UNKNOWN", [])  # Unknown cluster
_CLUSTER = Cluster.cluster()  # Working cluster



import warnings

import numpy as np
import tensorflow as tf
from keras import utils as k_utils
from keras.models import Model as _Model
from keras.utils.data_utils import GeneratorEnqueuer, OrderedEnqueuer
from keras.utils.generic_utils import Progbar

from medsegpy.utils import env

if env.is_tf2():
    from tensorflow.python.keras import callbacks as callbacks_module
    from tensorflow.python.keras.engine import data_adapter
    from tensorflow.python.keras.engine.training import concat
    from tensorflow.python.keras.utils import tf_utils
    from tensorflow.python.util import nest
else:
    concat = None
    callbacks_module = None
    data_adapter = None
    nest = None
    tf_utils = None

__all__ = ["Model"]


class Model(_Model):
    """MedSegPy implementation of the Keras :class:`Model`.

    In addition to the traditional :class:`Model` functionality in Keras, this
    class offers the option of returning batches of inputs, outputs, and
    predictions using :meth:`inference_generator`.

    All models implemented in medsegpy should use this class as the base class.
    """

    def inference_generator(
        self,
        generator,
        steps=None,
        max_queue_size=10,
        workers=1,
        use_multiprocessing=False,
        verbose=0,
    ):
        return self.inference_generator_static(
            self, generator, steps, max_queue_size, workers, use_multiprocessing, verbose
        )

    @classmethod
    def inference_generator_static(
        cls,
        model,
        generator,
        steps=None,
        max_queue_size=10,
        workers=1,
        use_multiprocessing=False,
        verbose=0,
    ):
        """Generates predictions for the input samples from a data generator
        and returns inputs, ground truth, and predictions.

        The generator should return the same kind of data as accepted by
        `predict_on_batch`.

        Arguments:
            model: The Keras model.
            generator: Generator yielding batches of input samples
                or an instance of Sequence (keras.utils.data_utils.Sequence)
                object in order to avoid duplicate data
                when using multiprocessing.
            steps: Total number of steps (batches of samples)
                to yield from `generator` before stopping.
                Optional for `Sequence`: if unspecified, will use
                the `len(generator)` as a number of steps.
            max_queue_size: Maximum size for the generator queue.
            workers: Integer. Maximum number of processes to spin up
                when using process based threading.
                If unspecified, `workers` will default to 1. If 0, will
                execute the generator on the main thread.
            use_multiprocessing: If `True`, use process based threading.
                Note that because
                this implementation relies on multiprocessing,
                you should not pass
                non picklable arguments to the generator
                as they can't be passed
                easily to children processes.
            verbose: verbosity mode, 0 or 1.

        Returns:
            Numpy array(s) of inputs, outputs, predictions.

        # Raises
            ValueError: In case the generator yields
                data in an invalid format.
        """
        # Issue #8: Remove this check when shuffle=True is supported.
        if (
            hasattr(generator, "shuffle")
            and isinstance(generator.shuffle, bool)
            and generator.shuffle
        ):
            raise ValueError(
                "Shuffling in generator is not supported. " "Set `generator.shuffle=False`."
            )

        if env.is_tf2():
            # TODO (TF2.X): Update when dataloaders migrate from keras.Sequence -> tf.data
            return cls._inference_generator_tf2(
                model=model,
                x=generator,
                steps=steps,
                max_queue_size=max_queue_size,
                workers=workers,
                use_multiprocessing=use_multiprocessing,
                verbose=verbose,
            )
        else:
            return model._inference_generator_tf1(
                generator=generator,
                steps=steps,
                max_queue_size=max_queue_size,
                workers=workers,
                use_multiprocessing=use_multiprocessing,
                verbose=verbose,
            )

    def _inference_generator_tf1(
        self,
        generator,
        steps=None,
        max_queue_size=10,
        workers=1,
        use_multiprocessing=False,
        verbose=0,
    ):
        """Inference generator for TensorFlow 1."""
        self._make_predict_function()

        steps_done = 0
        wait_time = 0.01
        all_outs = []
        all_xs = []
        all_ys = []
        is_sequence = isinstance(generator, k_utils.Sequence)
        if not is_sequence and use_multiprocessing and workers > 1:
            warnings.warn(
                UserWarning(
                    "Using a generator with `use_multiprocessing=True`"
                    " and multiple workers may duplicate your data."
                    " Please consider using the`keras.utils.data_utils.Sequence"
                    " class."
                )
            )
        if steps is None:
            if is_sequence:
                steps = len(generator)
            else:
                raise ValueError(
                    "`steps=None` is only valid for a generator"
                    " based on the `keras.utils.data_utils.Sequence` class."
                    " Please specify `steps` or use the"
                    " `keras.utils.data_utils.Sequence` class."
                )
        enqueuer = None

        try:
            if workers > 0:
                if is_sequence:
                    enqueuer = OrderedEnqueuer(generator, use_multiprocessing=use_multiprocessing)
                else:
                    enqueuer = GeneratorEnqueuer(
                        generator, use_multiprocessing=use_multiprocessing, wait_time=wait_time
                    )
                enqueuer.start(workers=workers, max_queue_size=max_queue_size)
                output_generator = enqueuer.get()
            else:
                if is_sequence:
                    output_generator = iter(generator)
                else:
                    output_generator = generator

            if verbose == 1:
                progbar = Progbar(target=steps)

            while steps_done < steps:
                generator_output = next(output_generator)
                x, y, x_raw = _extract_inference_inputs(generator_output)

                outs = self.predict_on_batch(x)
                xs = [x_raw if x_raw is not None else x]
                ys = [y]
                if not isinstance(outs, list):
                    outs = [outs]

                if not all_outs:
                    for _ in outs:
                        all_outs.append([])
                if not all_xs:
                    for _ in xs:
                        all_xs.append([])
                if not all_ys:
                    for _ in ys:
                        all_ys.append([])

                for i, out in enumerate(outs):
                    all_outs[i].append(out)
                for i, x in enumerate(xs):
                    all_xs[i].append(x)
                for i, y in enumerate(ys):
                    all_ys[i].append(y)
                steps_done += 1
                if verbose == 1:
                    progbar.update(steps_done)

        finally:
            if enqueuer is not None:
                enqueuer.stop()

        if len(all_outs) == 1:
            if steps_done == 1:
                return all_xs[0][0], all_ys[0][0], all_outs[0][0]
            else:
                return (
                    np.concatenate(all_xs[0]),
                    np.concatenate(all_ys[0]),
                    np.concatenate(all_outs[0]),
                )
        if steps_done == 1:
            return (
                [xs[0] for xs in all_xs],
                [ys[0] for ys in all_ys],
                [out[0] for out in all_outs],
            )
        else:
            return (
                [np.concatenate(xs) for xs in all_xs],
                [np.concatenate(ys) for ys in all_ys],
                [np.concatenate(out) for out in all_outs],
            )

    @staticmethod
    def _inference_generator_tf2(
        model: _Model,
        x,
        batch_size=None,
        verbose=0,
        steps=None,
        callbacks=None,
        max_queue_size=10,
        workers=1,
        use_multiprocessing=False,
    ):
        """Inference generator for TensorFlow 2."""
        outputs = []
        xs = []
        ys = []
        with model.distribute_strategy.scope():
            # Creates a `tf.data.Dataset` and handles batch and epoch iteration.
            data_handler = data_adapter.DataHandler(
                x=x,
                batch_size=batch_size,
                steps_per_epoch=steps,
                initial_epoch=0,
                epochs=1,
                max_queue_size=max_queue_size,
                workers=workers,
                use_multiprocessing=use_multiprocessing,
                model=model,
                steps_per_execution=model._steps_per_execution,
            )

            # Container that configures and calls `tf.keras.Callback`s.
            if not isinstance(callbacks, callbacks_module.CallbackList):
                callbacks = callbacks_module.CallbackList(
                    callbacks,
                    add_history=True,
                    add_progbar=verbose != 0,
                    model=model,
                    verbose=verbose,
                    epochs=1,
                    steps=data_handler.inferred_steps,
                )

            # predict_function = model.make_predict_function()
            model._predict_counter.assign(0)
            callbacks.on_predict_begin()
            for _, iterator in data_handler.enumerate_epochs():  # Single epoch.
                # iterator = peekable(iterator)
                with data_handler.catch_stop_iteration():
                    for step in data_handler.steps():
                        callbacks.on_predict_batch_begin(step)
                        batch_x, batch_y, batch_x_raw = _extract_inference_inputs(next(iterator))
                        # tmp_batch_outputs = predict_function(iterator)
                        tmp_batch_outputs = model.predict(batch_x)
                        if data_handler.should_sync:
                            context.async_wait()  # noqa: F821
                        batch_outputs = tmp_batch_outputs  # No error, now safe to assign.

                        if batch_x_raw is not None:
                            batch_x = batch_x_raw
                        for batch, running in zip(
                            [batch_x, batch_y, batch_outputs], [xs, ys, outputs]
                        ):
                            nest.map_structure_up_to(
                                batch, lambda x, batch_x: x.append(batch_x), running, batch
                            )

                        end_step = step + data_handler.step_increment
                        callbacks.on_predict_batch_end(end_step, {"outputs": batch_outputs})
                callbacks.on_predict_end()

            xs = [tf_utils.to_numpy_or_python_type(_x) for _x in xs]
            ys = [tf_utils.to_numpy_or_python_type(_y) for _y in ys]
            outputs = [tf_utils.to_numpy_or_python_type(_o) for _o in outputs]
            all_xs = nest.map_structure_up_to(batch_x, np.concatenate, xs)
            all_ys = nest.map_structure_up_to(batch_y, np.concatenate, ys)
            all_outputs = nest.map_structure_up_to(batch_outputs, np.concatenate, outputs)
            return all_xs, all_ys, all_outputs

            # all_xs = nest.map_structure_up_to(batch_x, concat, xs)
            # all_ys = nest.map_structure_up_to(batch_y, concat, ys)
            # all_outputs = nest.map_structure_up_to(batch_outputs, concat, outputs)
            # return (
            #     tf_utils.to_numpy_or_python_type(all_xs),
            #     tf_utils.to_numpy_or_python_type(all_ys),
            #     tf_utils.to_numpy_or_python_type(all_outputs),
            # )


def _extract_inference_inputs(inputs):
    def check_type(x):
        if env.is_tf2():
            return isinstance(x, tf.Tensor)
        else:
            return isinstance(x, np.ndarray)

    if isinstance(inputs, tuple):
        # Compatibility with the generators
        # used for training.
        if len(inputs) == 2:
            x, y = inputs
            x_raw = None
        elif len(inputs) == 3:
            x, y, x_raw = inputs
            assert check_type(x_raw)
        else:
            raise ValueError(
                "Output of generator should be "
                "a tuple `(x, y, x_raw)` "
                "or `(x, y)`. Found: " + str(inputs)
            )
    else:
        # Assumes a generator that only
        # yields inputs and targets (not sample weights).
        x, y = inputs
        x_raw = None

    return x, y, x_raw



import copy
import logging
import os
import time
from typing import Sequence, Union

import h5py
import numpy as np
from fvcore.common.file_io import PathManager

from medsegpy.config import Config
from medsegpy.data import MetadataCatalog

from .build import EVALUATOR_REGISTRY
from .evaluator import DatasetEvaluator
from .metrics import Metric, MetricsManager


def get_stats_string(manager: MetricsManager):
    """Returns formatted metrics manager summary string.

    Args:
        manager (MetricsManager): The manager whose results to format.

    Returns:
        str: A formatted string detailing manager results.
    """
    inference_runtimes = np.asarray(manager.runtimes)

    s = "============ Overall Summary ============\n"
    s += "%s\n" % manager.summary()
    s += "Inference time (Mean +/- Std. Dev.): " "{:0.2f} +/- {:0.2f} seconds.\n".format(
        np.mean(inference_runtimes), np.std(inference_runtimes)
    )
    return s


@EVALUATOR_REGISTRY.register()
class SemSegEvaluator(DatasetEvaluator):
    """Evaluator for semantic segmentation-related tasks."""

    def __init__(
        self,
        dataset_name: str,
        cfg: Config,
        output_folder: str = None,
        save_raw_data: bool = False,
        stream_evaluation: bool = True,
    ):
        """
        Args:
            dataset_name (str): name of the dataset to be evaluated.
            cfg:
            output_folder (str): an output directory to dump results.
            save_raw_data (:obj:`bool`, optional): Save probs, labels, ground
                truth masks to h5 file.
            stream_evaluation (:obj:`bool`, optional): If `True`, evaluates
                data as it comes in to avoid holding too many objects in memory.
        """
        self._config = cfg
        self._dataset_name = dataset_name
        self._output_folder = (
            output_folder if output_folder else os.path.join(cfg.OUTPUT_DIR, "test_results")
        )
        PathManager.mkdirs(self._output_folder)
        self._num_classes = cfg.get_num_classes()
        self._ignore_label = 0

        self._logger = logging.getLogger(__name__)

        meta = MetadataCatalog.get(dataset_name)
        self._meta = meta
        cat_ids = cfg.CATEGORIES
        contiguous_id_map = meta.get("category_id_to_contiguous_id")
        if type(contiguous_id_map) is dict:
            contiguous_id_map = contiguous_id_map[cfg.PRIMARY_TASK]
        contiguous_ids = [
            contiguous_id_map[tuple(x) if isinstance(x, list) else x] for x in cat_ids
        ]
        categories = meta.get("categories")
        if type(categories) is dict:
            categories = categories[cfg.PRIMARY_TASK]
        categories = [categories[c_id] for c_id in contiguous_ids]
        self.spacing = meta.get("spacing", None)

        self._categories = categories

        self._metrics_manager = None
        self._predictions = None
        self._scan_cnt = 0
        self._results_str = ""

        self._save_raw_data = save_raw_data
        self.stream_evaluation = stream_evaluation

        self._output_activation = cfg.LOSS[1]
        self._output_includes_background = cfg.INCLUDE_BACKGROUND

    def _get_metrics(self) -> Sequence[Union[str, Metric]]:
        """"""
        return self._config.TEST_METRICS

    def reset(self):
        self._metrics_manager = MetricsManager(
            class_names=self._categories, metrics=self._get_metrics()
        )
        self._predictions = []
        self._scan_cnt = 0
        self._results_str = ""

    def process(self, inputs, outputs):
        """
        See :class:`DatasetEvaluator` in evaluator.py
        for argument details.
        """
        output_activation = self._output_activation
        includes_bg = self._output_includes_background

        for input, output in zip(inputs, outputs):
            # Copy because we may modify below
            input = copy.deepcopy(input)
            output = copy.deepcopy(output)

            y_pred = output["y_pred"]

            if output_activation == "sigmoid":
                labels = (y_pred > 0.5).astype(np.uint8)
            elif output_activation == "softmax":
                labels = np.zeros_like(y_pred, dtype=np.uint8)
                l_argmax = np.argmax(y_pred, axis=-1)
                for c in range(labels.shape[-1]):
                    labels[l_argmax == c, c] = 1
                labels = labels.astype(np.uint)
            else:
                raise ValueError("output activation {} not supported".format(output_activation))

            # background is always excluded from analysis
            if includes_bg:
                y_true = output["y_true"][..., 1:]
                y_pred = output["y_pred"][..., 1:]
                labels = labels[..., 1:]
                # if y_true.ndim == 3:
                #     y_true = y_true[..., np.newaxis]
                #     y_pred = y_pred[..., np.newaxis]
                #     labels = labels[..., np.newaxis]
                output["y_true"] = y_true
                output["y_pred"] = y_pred

            time_elapsed = output["time_elapsed"]
            if self.stream_evaluation:
                self.eval_single_scan(input, output, labels, time_elapsed)
            else:
                self._predictions.append((input, output, labels, time_elapsed))

    def eval_single_scan(self, input, output, labels, time_elapsed):
        metrics_manager = self._metrics_manager
        spacing = input["scan_spacing"] if "scan_spacing" in input else self.spacing
        logger = self._logger
        save_raw_data = self._save_raw_data
        output_dir = self._output_folder

        self._scan_cnt += 1
        scan_cnt = self._scan_cnt
        scan_id = input["scan_id"]
        y_true: np.ndarray = output["y_true"]
        x: np.ndarray = np.squeeze(input["x"])

        metrics_kwargs = {"spacing": spacing} if spacing is not None else {}

        summary = metrics_manager(
            scan_id, y_true=y_true, y_pred=labels, x=x, runtime=time_elapsed, **metrics_kwargs
        )

        logger_info_str = "Scan #{:03d} (name = {}, {:0.2f}s) = {}".format(
            scan_cnt, scan_id, time_elapsed, summary
        )
        self._results_str = self._results_str + logger_info_str + "\n"
        logger.info(logger_info_str)

        if output_dir and save_raw_data:
            save_name = "{}/{}.pred".format(output_dir, scan_id)
            with h5py.File(save_name, "w") as h5f:
                h5f.create_dataset("probs", data=output["y_pred"])
                h5f.create_dataset("labels", data=labels)

    def evaluate(self):
        """Evaluates popular medical segmentation metrics specified in config.

        * Evaluate on popular medical segmentation metrics. For supported
          segmentation metrics, see :class:`MetricsManager`.
        * Save overlay images.
        * Save probability predictions.

        Note, by default coefficient of variation (CV) is calculated as a
        root-mean-squared quantity rather than mean.
        """
        output_dir = self._output_folder
        logger = self._logger

        if self._predictions:
            for input, output, labels, time_elapsed in self._predictions:
                self.eval_single_scan(input, output, labels, time_elapsed)

        results_str = self._results_str
        stats_string = get_stats_string(self._metrics_manager)
        logger.info("--" * 20)
        logger.info("\n" + stats_string)
        logger.info("--" * 20)
        if output_dir:
            test_results_summary_path = os.path.join(output_dir, "results.txt")

            # Write details to test file
            with open(test_results_summary_path, "w+") as f:
                f.write("Results generated on %s\n" % time.strftime("%X %x %Z"))
                f.write("Weights Loaded: %s\n" % os.path.basename(self._config.TEST_WEIGHT_PATH))

                f.write("--" * 20)
                f.write("\n")
                f.write(results_str)
                f.write("--" * 20)
                f.write("\n")
                f.write(stats_string)

        # df = self._metrics_manager.data_frame()
        # df.to_csv(os.path.join(output_dir, "metrics.csv"), header=True, index=True)  # noqa

        df = self._metrics_manager.data()["scan_data"]
        df.to_csv(os.path.join(output_dir, "metrics.csv"), header=True, index=True)

        # TODO: Convert segmentation metrics to valid results matrix.
        return {}



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import functools
import logging
import os
import sys
import time
from collections import Counter

from fvcore.common.file_io import PathManager
from tabulate import tabulate
from termcolor import colored

logging.captureWarnings(True)


class _ColorfulFormatter(logging.Formatter):
    def __init__(self, *args, **kwargs):
        self._root_name = kwargs.pop("root_name") + "."
        self._abbrev_name = kwargs.pop("abbrev_name", "")
        if len(self._abbrev_name):
            self._abbrev_name = self._abbrev_name + "."
        super(_ColorfulFormatter, self).__init__(*args, **kwargs)

    def formatMessage(self, record):
        record.name = record.name.replace(self._root_name, self._abbrev_name)
        log = super(_ColorfulFormatter, self).formatMessage(record)
        if record.levelno == logging.WARNING:
            prefix = colored("WARNING", "red", attrs=["blink"])
        elif record.levelno == logging.ERROR or record.levelno == logging.CRITICAL:
            prefix = colored("ERROR", "red", attrs=["blink", "underline"])
        else:
            return log
        return prefix + " " + log


@functools.lru_cache()  # so that calling setup_logger multiple times won't add many handlers  # noqa
def setup_logger(output=None, distributed_rank=0, *, color=True, name="medsegpy", abbrev_name=None):
    """
    Initialize the detectron2 logger and set its verbosity level to "INFO".

    Args:
        output (str): a file name or a directory to save log. If None, will not
            save log file. If ends with ".txt" or ".log", assumed to be a file
            name. Otherwise, logs will be saved to `output/log.txt`.
        name (str): the root module name of this logger
        abbrev_name (str): an abbreviation of the module, to avoid long names in
            logs. Set to "" to not log the root module in logs.
            By default, will abbreviate "detectron2" to "d2" and leave other
            modules unchanged.

    Returns:
        logging.Logger: a logger
    """
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False
    if abbrev_name is None:
        abbrev_name = name

    plain_formatter = logging.Formatter(
        "[%(asctime)s] %(name)s %(levelname)s: %(message)s", datefmt="%m/%d %H:%M:%S"
    )
    # stdout logging: master only
    if distributed_rank == 0:
        ch = logging.StreamHandler(stream=sys.stdout)
        ch.setLevel(logging.DEBUG)
        if color:
            formatter = _ColorfulFormatter(
                colored("[%(asctime)s %(name)s]: ", "green") + "%(message)s",
                datefmt="%m/%d %H:%M:%S",
                root_name=name,
                abbrev_name=str(abbrev_name),
            )
        else:
            formatter = plain_formatter
        ch.setFormatter(formatter)
        logger.addHandler(ch)

    # file logging: all workers
    if output is not None:
        if output.endswith(".txt") or output.endswith(".log"):
            filename = output
        else:
            filename = os.path.join(output, "log.txt")
        if distributed_rank > 0:
            filename = filename + ".rank{}".format(distributed_rank)
        PathManager.mkdirs(os.path.dirname(filename))

        fh = logging.StreamHandler(_cached_log_stream(filename))
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(plain_formatter)
        logger.addHandler(fh)

    return logger


# cache the opened file object, so that different calls to `setup_logger`
# with the same file name can safely write to the same file.
@functools.lru_cache(maxsize=None)
def _cached_log_stream(filename):
    return PathManager.open(filename, "a")


"""
Below are some other convenient logging methods.
They are mainly adopted from
https://github.com/abseil/abseil-py/blob/master/absl/logging/__init__.py
"""


def _find_caller():
    """
    Returns:
        str: module name of the caller
        tuple: a hashable key to be used to identify different callers
    """
    frame = sys._getframe(2)
    while frame:
        code = frame.f_code
        if os.path.join("utils", "logger.") not in code.co_filename:
            mod_name = frame.f_globals["__name__"]
            if mod_name == "__main__":
                mod_name = "detectron2"
            return mod_name, (code.co_filename, frame.f_lineno, code.co_name)
        frame = frame.f_back


_LOG_COUNTER = Counter()
_LOG_TIMER = {}


def log_first_n(lvl, msg, n=1, *, name=None, key="caller"):
    """
    Log only for the first n times.

    Args:
        lvl (int): the logging level
        msg (str):
        n (int):
        name (str): name of the logger to use. Will use the caller's module by
            default.
        key (str or tuple[str]): the string(s) can be one of "caller" or
            "message", which defines how to identify duplicated logs.
            For example, if called with `n=1, key="caller"`, this function
            will only log the first call from the same caller, regardless of
            the message content.
            If called with `n=1, key="message"`, this function will log the
            same content only once, even if they are called from different
            places.
            If called with `n=1, key=("caller", "message")`, this function
            will not log only if the same caller has logged the same message
            before.
    """
    if isinstance(key, str):
        key = (key,)
    assert len(key) > 0

    caller_module, caller_key = _find_caller()
    hash_key = ()
    if "caller" in key:
        hash_key = hash_key + caller_key
    if "message" in key:
        hash_key = hash_key + (msg,)

    _LOG_COUNTER[hash_key] += 1
    if _LOG_COUNTER[hash_key] <= n:
        logging.getLogger(name or caller_module).log(lvl, msg)


def log_every_n(lvl, msg, n=1, *, name=None):
    """
    Log once per n times.

    Args:
        lvl (int): the logging level
        msg (str):
        n (int):
        name (str): name of the logger to use. Will use the caller's module by
            default.
    """
    caller_module, key = _find_caller()
    _LOG_COUNTER[key] += 1
    if n == 1 or _LOG_COUNTER[key] % n == 1:
        logging.getLogger(name or caller_module).log(lvl, msg)


def log_every_n_seconds(lvl, msg, n=1, *, name=None):
    """
    Log no more than once per n seconds.

    Args:
        lvl (int): the logging level
        msg (str):
        n (int):
        name (str): name of the logger to use. Will use the caller's module by
            default.
    """
    caller_module, key = _find_caller()
    last_logged = _LOG_TIMER.get(key, None)
    current_time = time.time()
    if last_logged is None or current_time - last_logged >= n:
        logging.getLogger(name or caller_module).log(lvl, msg)
        _LOG_TIMER[key] = current_time


def create_small_table(small_dict):
    """
    Create a small table using the keys of small_dict as headers. This is only
    suitable for small dictionaries.

    Args:
        small_dict (dict): a result dictionary of only a few items.

    Returns:
        str: the table as a string.
    """
    keys, values = tuple(zip(*small_dict.items()))
    table = tabulate(
        [values],
        headers=keys,
        tablefmt="pipe",
        floatfmt=".3f",
        stralign="center",
        numalign="center",
    )
    return table



"""
Create K bins for K-fold cross-validation

Data is stored in Pickle format
"""

import argparse

from .cv_util import DATASET_NAME, CrossValidationFileGenerator

DATA_PATHS = [defaults.TRAIN_PATH, defaults.VALID_PATH, defaults.TEST_PATH]

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Generate bins for k-fold cross-validation")

    parser.add_argument(
        "-k",
        "--k_bins",
        metavar="K",
        nargs=1,
        type=int,
        help="Number of bins to create for k-fold cross-validation",
    )

    parser.add_argument(
        "--filepaths", metavar="F", nargs="?", default=DATA_PATHS, help="Filepaths to data"
    )

    parser.add_argument(
        "--tag",
        metavar="T",
        nargs="?",
        default=DATASET_NAME,
        help="Tag uniquely identifying dataset used",
    )

    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Overwrite existing file with same name",
        dest="overwrite",
    )

    args = parser.parse_args()

    k_bins = args.k_bins[0]
    filepaths = args.filepaths
    tag = args.tag
    overwrite = args.overwrite

    cv_generator = CrossValidationFileGenerator(k_bins, filepaths, tag, overwrite=overwrite)
